import logging

import celery_pubsub
from celery import shared_task
from django.db.models import Q
from web3.exceptions import TransactionNotFound

from hub20.apps.blockchain.client import make_web3
from hub20.apps.blockchain.models import (
    BaseEthereumAccount,
    Chain,
    Transaction,
    TransactionDataRecord,
    Web3Provider,
)

from . import signals
from .models import EthereumToken, TokenList

logger = logging.getLogger(__name__)


@shared_task
def import_token_list(url, description=None):
    token_list_data = TokenList.fetch(url)
    TokenList.make(url, token_list_data, description=description)


@shared_task
def record_token_transfers(chain_id, event_data, provider_url):
    token_address = event_data.address
    token = EthereumToken.objects.filter(chain_id=chain_id, address=token_address).first()

    if not token:
        return

    sender = event_data.args._from
    recipient = event_data.args._to

    if not BaseEthereumAccount.objects.filter(Q(address=sender) | Q(address=recipient)).exists():
        return

    try:
        provider = Web3Provider.objects.get(url=provider_url)
        w3 = make_web3(provider=provider)

        tx_data = w3.eth.get_transaction(event_data.transactionHash)
        tx_receipt = w3.eth.get_transaction_receipt(event_data.transactionHash)
        block_data = w3.eth.get_block(tx_receipt.blockHash)
        amount = token.from_wei(event_data.args._value)

        TransactionDataRecord.make(chain_id=chain_id, tx_data=tx_data)
        tx = Transaction.make(chain_id=chain_id, block_data=block_data, tx_receipt=tx_receipt)
    except TransactionNotFound:
        logger.warning(f"Failed to get transaction {event_data.transactionHash.hex()}")
        return

    for account in BaseEthereumAccount.objects.filter(address=sender):
        account.transactions.add(tx)
        signals.outgoing_transfer_mined.send(
            sender=Transaction,
            account=account,
            transaction=tx,
            amount=amount,
            address=recipient,
        )

    for account in BaseEthereumAccount.objects.filter(address=recipient):
        account.transactions.add(tx)
        signals.incoming_transfer_mined.send(
            sender=Transaction,
            account=account,
            transaction=tx,
            amount=amount,
            address=sender,
        )


@shared_task
def check_eth_transfers(chain_id, block_data, provider_url):
    addresses = BaseEthereumAccount.objects.values_list("address", flat=True)

    txs = [
        t
        for t in block_data["transactions"]
        if t.value > 0 and (t["to"] in addresses or t["from"] in addresses)
    ]

    if not txs:
        return

    chain = Chain.active.get(id=chain_id)
    provider = Web3Provider.objects.get(url=provider_url)
    w3 = make_web3(provider=provider)

    assert chain == provider.chain, f"{provider.hostname} not connected to {chain.name}"

    native_token = EthereumToken.make_native(chain=chain)

    for transaction_data in txs:
        sender = transaction_data["from"]
        recipient = transaction_data["to"]

        amount = native_token.from_wei(transaction_data.value)

        transaction_receipt = w3.eth.get_transaction_receipt(transaction_data.hash)
        tx = Transaction.make(
            chain_id=chain_id,
            block_data=block_data,
            tx_receipt=transaction_receipt,
        )

        for account in BaseEthereumAccount.objects.filter(address=sender):
            account.transactions.add(tx)
            signals.outgoing_transfer_mined.send(
                sender=Transaction,
                account=account,
                amount=amount,
                transaction=tx,
                address=recipient,
            )

        for account in BaseEthereumAccount.objects.filter(address=recipient):
            account.transactions.add(tx)
            signals.incoming_transfer_mined.send(
                sender=Transaction,
                account=account,
                amount=amount,
                transaction=tx,
                address=sender,
            )


@shared_task
def check_pending_transaction_for_eth_transfer(chain_id, transaction_data):
    chain = Chain.actve.get(id=chain_id)

    sender = transaction_data["from"]
    recipient = transaction_data["to"]

    is_native_token_transfer = transaction_data.value != 0

    if not is_native_token_transfer:
        return

    native_token = EthereumToken.make_native(chain=chain)
    amount = native_token.from_wei(transaction_data.value)

    for account in BaseEthereumAccount.objects.filter(address=sender):
        tx_data = TransactionDataRecord.make(tx_data=transaction_data, chain_id=chain_id)

        signals.outgoing_transfer_broadcast.send(
            sender=TransactionDataRecord,
            account=account,
            amount=amount,
            transaction_data=tx_data,
        )

    for account in BaseEthereumAccount.objects.filter(address=recipient):
        tx_data = TransactionDataRecord.make(tx_data=transaction_data, chain_id=chain_id)

        signals.incoming_transfer_broadcast.send(
            sender=EthereumToken,
            account=account,
            amount=amount,
            transaction_data=tx_data,
        )


@shared_task
def check_pending_erc20_transfer_event(chain_id, event_data, provider_url):
    try:
        token = EthereumToken.objects.get(chain_id=chain_id, address=event_data.address)
    except EthereumToken.DoesNotExist:
        return

    sender = event_data.args._from
    recipient = event_data.args._to

    if not BaseEthereumAccount.objects.filter(Q(address=sender) | Q(address=recipient)).exists():
        return

    amount = token.from_wei(event_data.args._value)

    try:
        provider = Web3Provider.objects.get(url=provider_url)
        w3 = make_web3(provider=provider)
        transaction_data = w3.eth.get_transaction(event_data.transactionHash)
    except TransactionNotFound:
        logger.warning(f"Failed to get transaction data {event_data.transactionHash.hex()}")
        return

    for account in BaseEthereumAccount.objects.filter(address=sender):
        tx_data = TransactionDataRecord.make(tx_data=transaction_data, chain_id=chain_id)
        signals.outgoing_transfer_broadcast.send(
            sender=TransactionDataRecord,
            account=account,
            amount=amount,
            transaction_data=tx_data,
        )

    for account in BaseEthereumAccount.objects.filter(address=recipient):
        tx_data = TransactionDataRecord.make(tx_data=transaction_data, chain_id=chain_id)
        signals.incoming_transfer_broadcast.send(
            sender=TransactionDataRecord,
            account=account,
            amount=amount,
            transaction_data=tx_data,
        )


celery_pubsub.subscribe("blockchain.event.token_transfer", record_token_transfers)
celery_pubsub.subscribe("blockchain.mined.block", check_eth_transfers)
celery_pubsub.subscribe(
    "blockchain.broadcast.transaction", check_pending_transaction_for_eth_transfer
)
celery_pubsub.subscribe(
    "blockchain.event.token_transfer.broadcast", check_pending_erc20_transfer_event
)



from django.dispatch import Signal

raiden_payment_received = Signal(["payment"])
raiden_payment_sent = Signal(["payment"])
service_deposit_sent = Signal(["transaction", "raiden", "amount", "contract_address"])



import pytest
from django.test import TestCase

from hub20.apps.core.factories import Erc20TokenPaymentOrderFactory, TransferFactory
from hub20.apps.core.models import (
    BlockchainPaymentRoute,
    PaymentConfirmation,
    PaymentOrder,
    RaidenPaymentRoute,
    Transfer,
    TransferExecution,
)
from hub20.apps.ethereum_money.tests.base import add_token_to_account
from hub20.apps.raiden.factories import ChannelFactory


@pytest.mark.django_db(transaction=True)
class BaseTestCase(TestCase):
    pass


class PaymentOrderManagerTestCase(BaseTestCase):
    def setUp(self):
        self.raiden_channel = ChannelFactory()
        self.order = Erc20TokenPaymentOrderFactory(
            currency=self.raiden_channel.token_network.token
        )

    def test_order_has_blockchain_route(self):
        self.assertTrue(PaymentOrder.objects.with_blockchain_route().exists())

    def test_order_has_raiden_route(self):
        self.assertTrue(RaidenPaymentRoute.objects.exists())
        with_route = PaymentOrder.objects.with_raiden_route()
        self.assertTrue(with_route.exists())

    def test_order_with_no_payment_is_open(self):
        self.assertTrue(PaymentOrder.objects.unpaid().filter(id=self.order.id).exists())
        self.assertFalse(PaymentOrder.objects.paid().filter(id=self.order.id).exists())

    def test_order_with_partial_payment_is_open(self):
        route = BlockchainPaymentRoute.objects.filter(deposit=self.order).first()
        partial_payment_amount = self.order.as_token_amount * 0.5
        tx = add_token_to_account(route.account, partial_payment_amount)
        PaymentConfirmation.objects.create(payment=tx.blockchainpayment)
        self.assertTrue(BlockchainPaymentRoute.objects.open().exists())

    def test_order_with_unconfirmed_payment_is_open(self):
        route = BlockchainPaymentRoute.objects.filter(deposit=self.order).first()
        add_token_to_account(route.account, self.order.as_token_amount)
        self.assertTrue(BlockchainPaymentRoute.objects.open().exists())

    def test_order_with_multiple_payments_is_not_open(self):
        route = BlockchainPaymentRoute.objects.filter(deposit=self.order).first()
        partial_payment_amount = self.order.as_token_amount * 0.5
        first_tx = add_token_to_account(route.account, partial_payment_amount)
        PaymentConfirmation.objects.create(payment=first_tx.blockchainpayment)
        self.assertTrue(BlockchainPaymentRoute.objects.open().exists())

        second_tx = add_token_to_account(route.account, partial_payment_amount)
        PaymentConfirmation.objects.create(payment=second_tx.blockchainpayment)
        self.assertFalse(BlockchainPaymentRoute.objects.open().exists())

    def test_order_with_confirmed_payment_is_not_open(self):
        route = BlockchainPaymentRoute.objects.filter(deposit=self.order).first()
        tx = add_token_to_account(route.account, self.order.as_token_amount)
        PaymentConfirmation.objects.create(payment=tx.blockchainpayment)
        self.assertFalse(BlockchainPaymentRoute.objects.open().exists())


class TransferManagerTestCase(BaseTestCase):
    def test_pending_query_manager(self):
        self.transfer = TransferFactory()

        self.assertTrue(Transfer.pending.exists())

        # Executed transfers are no longer pending
        TransferExecution.objects.create(transfer=self.transfer)
        self.assertTrue(Transfer.executed.exists())
        self.assertFalse(Transfer.pending.exists())

        # Another transfer shows up, and already confirmed transfers are out
        another_transfer = TransferFactory()
        self.assertTrue(Transfer.pending.exists())
        self.assertEqual(Transfer.pending.count(), 1)
        self.assertEqual(Transfer.pending.first(), another_transfer)


__all__ = ["PaymentOrderManagerTestCase", "TransferManagerTestCase"]



import logging

import celery_pubsub
from celery import shared_task
from web3.exceptions import TransactionNotFound

from hub20.apps.blockchain.client import make_web3
from hub20.apps.blockchain.models import Transaction, TransactionDataRecord, Web3Provider
from hub20.apps.ethereum_money.client import get_account_balance
from hub20.apps.ethereum_money.models import EthereumTokenAmount
from hub20.apps.raiden import models
from hub20.apps.raiden.client.blockchain import make_service_deposit
from hub20.apps.raiden.client.node import RaidenClient
from hub20.apps.raiden.exceptions import InsufficientBalanceError

logger = logging.getLogger(__name__)


@shared_task
def check_token_network_channel_events(chain_id, event_data, provider_url):
    token_network = models.TokenNetwork.objects.filter(address=event_data.address).first()

    if not token_network:
        return

    participants = (event_data.args.participant1, event_data.args.participant2)
    channel_identifier = event_data.args.channel_identifier
    channel, _ = token_network.channels.get_or_create(
        identifier=channel_identifier, participant_addresses=participants
    )

    try:
        provider = Web3Provider.objects.get(url=provider_url)
        w3 = make_web3(provider=provider)
        tx_hash = event_data.transactionHash

        tx_data = w3.eth.get_transaction(tx_hash)
        TransactionDataRecord.make(chain_id=chain_id, tx_data=tx_data)
        block_data = w3.eth.get_block(tx_data.blockHash)
        tx_receipt = w3.eth.get_transaction_receipt(tx_hash)
        tx = Transaction.make(chain_id=chain_id, block_data=block_data, tx_receipt=tx_receipt)

        token_network.events.get_or_create(channel=channel, transaction=tx, name=event_data.event)
    except TransactionNotFound:
        logger.warning(f"Failed to get information for Tx {event_data.transactionHash.hex()}")


@shared_task
def check_order_results(chain_id, block_data, provider_url):
    tx_hashes = [tx["hash"] for tx in block_data["transactions"]]
    open_order = models.RaidenManagementOrder.objects.filter(
        result__transaction__isnull=True, transaction_hash__in=tx_hashes
    ).first()

    if not open_order:
        return

    provider = Web3Provider.objects.get(url=provider_url)
    w3 = make_web3(provider=provider)

    try:
        transaction_receipt = w3.eth.get_transaction_receipt(open_order.transaction_hash)
    except TransactionNotFound:
        return

    tx = Transaction.make(
        chain_id=chain_id,
        block_data=block_data,
        tx_receipt=transaction_receipt,
    )

    successful = bool(transaction_receipt.status)

    model_class = (
        models.RaidenManagerOrderResult if successful else models.RaidenManagementOrderError
    )

    return model_class.objects.create(order=open_order, transaction=tx)


@shared_task
def make_udc_deposit(order_id: int):
    order = models.UserDepositContractOrder.objects.filter(id=order_id).first()

    if not order:
        logging.warning(f"UDC Order {order_id} not found")
        return

    w3 = make_web3(provider=order.currency.chain.provider)
    token_amount = order.as_token_amount

    try:
        make_service_deposit(w3=w3, account=order.raiden, amount=token_amount)
    except InsufficientBalanceError as exc:
        return models.RaidenManagementOrderError.objects.create(order=order, message=str(exc))


@shared_task
def make_channel_deposit(order_id: int):
    order = models.ChannelDepositOrder.objects.filter(id=order_id).first()

    if not order:
        logger.warning(f"Channel Deposit Order {order_id} not found")
        return

    w3 = make_web3(provider=order.channel.token.chain.provider)

    client = RaidenClient(raiden_account=order.raiden)
    token_amount = EthereumTokenAmount(currency=order.channel.token, amount=order.amount)

    chain_balance = get_account_balance(
        w3=w3, token=order.channel.token, address=order.raiden.address
    )

    if chain_balance < token_amount:
        logger.warning(f"Insufficient balance {chain_balance.formatted} to deposit on channel")
        return

    client.make_channel_deposit(order.channel, token_amount)


@shared_task
def make_channel_withdraw(order_id: int):
    order = models.ChannelWithdrawOrder.objects.filter(id=order_id).first()

    if not order:
        logger.warning(f"Channel Withdraw Order {order_id} not found")
        return

    client = RaidenClient(raiden_account=order.raiden)
    token_amount = EthereumTokenAmount(currency=order.channel.token, amount=order.amount)
    channel_balance = order.channel.balance_amount

    if channel_balance < token_amount:
        logger.warning(f"Insufficient balance {channel_balance.formatted} to withdraw")
        return

    client.make_channel_withdraw(order.channel, token_amount)


@shared_task
def join_token_network(order_id: int):
    order = models.JoinTokenNetworkOrder.objects.filter(id=order_id).first()

    if not order:
        logger.warning(f"Join Token Network Order {order_id} not found")

    client = RaidenClient(raiden_account=order.raiden)
    token_amount = EthereumTokenAmount(currency=order.token_network.token, amount=order.amount)

    w3 = make_web3(provider=order.token_network.token.chain.provider)

    chain_balance = get_account_balance(
        w3=w3, token=order.token_network.token, address=order.raiden.address
    )

    if chain_balance < token_amount:
        logger.warning(
            f"Balance {chain_balance.formatted} smaller than request to join token network"
        )
        return

    client.join_token_network(token_network=order.token_network, amount=token_amount)


@shared_task
def leave_token_network(order_id: int):
    order = models.LeaveTokenNetworkOrder.objects.filter(id=order_id).first()

    if not order:
        logger.warning(f"Leave Token Network Order {order_id} not found")
        return

    client = RaidenClient(raiden_account=order.raiden)
    client.leave_token_network(token_network=order.token_network)


celery_pubsub.subscribe(
    "blockchain.event.token_network_channel_opened.mined", check_token_network_channel_events
)
celery_pubsub.subscribe(
    "blockchain.event.token_network_channel_closed.mined", check_token_network_channel_events
)
celery_pubsub.subscribe("blockchain.transaction.mined", check_order_results)



from django.dispatch import Signal

incoming_transfer_broadcast = Signal(["account", "amount", "transaction_data"])
incoming_transfer_mined = Signal(["account", "transaction", "amount", "address"])
outgoing_transfer_broadcast = Signal(["account", "amount", "transaction_data"])
outgoing_transfer_mined = Signal(["account", "transaction", "amount", "address"])


__all__ = [
    "incoming_transfer_broadcast",
    "incoming_transfer_mined",
    "outgoing_transfer_broadcast",
    "outgoing_transfer_mined",
]



from rest_framework.exceptions import APIException


class EthereumNodeError(APIException):
    status_code = 502
    default_detail = "Ethereum Node Error"
    default_code = "ethereum_node_error"


class EthereumNodeUnavailable(APIException):
    status_code = 503
    default_detail = "Ethereum node is unavailable, please try again later"
    default_code = "ethereum_node_unavailable"


class EthereumNodeOutOfSync(APIException):
    status_code = 503
    default_detail = "Ethereum node is still syncing, please try again later"
    default_code = "ethereum_node_out_of_sync"


class HubServiceOutOfSync(APIException):
    status_code = 503
    default_detail = "Hub node is out of sync with ethereum network, please try again later"
    default_code = "hub_service_out_of_sync"


class Web3TransactionError(Exception):
    pass


class Web3UnsupportedMethod(Exception):
    pass



import datetime
import json
import logging
from typing import Optional
from urllib.parse import urlparse

from django.contrib.postgres.fields import ArrayField, HStoreField
from django.db import models
from django.db.models import Max
from django.db.transaction import atomic
from django.utils import timezone
from model_utils.managers import InheritanceManager, QueryManager
from web3 import Web3
from web3.datastructures import AttributeDict

from .app_settings import START_BLOCK_NUMBER
from .fields import EthereumAddressField, HexField, Web3ProviderURLField

logger = logging.getLogger(__name__)


def serialize_web3_data(data: AttributeDict):
    return json.loads(Web3.toJSON(data))


class Chain(models.Model):
    id = models.PositiveBigIntegerField(primary_key=True)
    name = models.CharField(max_length=128, default="EVM-compatible network")
    is_mainnet = models.BooleanField(default=True)
    highest_block = models.PositiveIntegerField()
    objects = models.Manager()
    active = QueryManager(providers__is_active=True)

    @property
    def provider(self):
        return self.providers.filter(is_active=True).first()

    @property
    def synced(self):
        return self.provider.synced and self.provider.is_active

    def __str__(self):
        return f"{self.name} ({self.id})"

    class Meta:
        ordering = ("id",)


class Block(models.Model):
    hash = HexField(max_length=64, primary_key=True)
    chain = models.ForeignKey(Chain, on_delete=models.CASCADE, related_name="blocks")
    number = models.PositiveIntegerField(db_index=True)
    base_fee_per_gas = models.PositiveBigIntegerField(null=True)
    timestamp = models.DateTimeField()
    parent_hash = HexField(max_length=64)
    uncle_hashes = ArrayField(HexField(max_length=64))

    def __str__(self) -> str:
        return f"{self.hash} #{self.number}"

    @property
    def parent(self):
        return self.__class__.objects.filter(hash=self.parent_hash).first()

    @property
    def uncles(self):
        return self.__class__.objects.filter(hash__in=self.uncle_hashes)

    @property
    def confirmations(self) -> int:
        return self.chain.highest_block - self.number

    @classmethod
    def make(cls, block_data, chain_id: int):
        block_time = datetime.datetime.fromtimestamp(block_data.timestamp)
        block, _ = cls.objects.update_or_create(
            chain_id=chain_id,
            hash=block_data.hash,
            defaults={
                "number": block_data.number,
                "timestamp": timezone.make_aware(block_time),
                "parent_hash": block_data.parentHash,
                "uncle_hashes": block_data.uncles,
                "base_fee_per_gas": getattr(block_data, "baseFeePerGas", None),
            },
        )
        return block

    @classmethod
    def get_latest_block_number(cls, qs):
        return qs.aggregate(latest=Max("number")).get("latest") or START_BLOCK_NUMBER

    class Meta:
        unique_together = ("chain", "hash", "number")


class AbstractTransactionRecord(models.Model):
    hash = HexField(max_length=64, db_index=True)
    from_address = EthereumAddressField(db_index=True)
    to_address = EthereumAddressField(db_index=True)

    def __str__(self) -> str:
        return f"{self.__class__.__name__} {self.hash}"

    class Meta:
        abstract = True


class TransactionDataRecord(AbstractTransactionRecord):
    """
    Transaction data records do not represent transactions by themselves
    """

    chain = models.ForeignKey(Chain, related_name="transaction_data", on_delete=models.CASCADE)
    data = HStoreField()

    @classmethod
    def make(cls, chain_id: int, tx_data: AttributeDict, force=False):
        action = cls.objects.update_or_create if force else cls.objects.get_or_create
        data, _ = action(
            chain_id=chain_id,
            hash=tx_data.hash,
            defaults={
                "from_address": tx_data["from"],
                "to_address": tx_data.to,
                "data": serialize_web3_data(tx_data),
            },
        )
        return data

    class Meta:
        unique_together = ("hash", "chain")


class Transaction(AbstractTransactionRecord):
    """
    Transaction models record the receipt, and should store
    all information related to a mined transaction
    """

    block = models.ForeignKey(Block, related_name="transactions", on_delete=models.CASCADE)
    receipt = HStoreField()

    def __str__(self) -> str:
        return f"Tx {self.hash}"

    @property
    def gas_used(self) -> int:
        return int(self.receipt.get("gasUsed", 0))

    @property
    def gas_price(self) -> int:
        return int(self.receipt.get("effectiveGasPrice", 0))

    @property
    def gas_fee(self) -> int:
        return self.gas_used * self.gas_price

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["hash", "block"], name="unique_tx_per_block"),
        ]

    @classmethod
    def make(
        cls,
        chain_id: int,
        tx_receipt: AttributeDict,
        block_data: AttributeDict,
        force=False,
    ):
        block = Block.make(chain_id=chain_id, block_data=block_data)
        action = cls.objects.update_or_create if force else cls.objects.get_or_create
        tx, _ = action(
            block=block,
            hash=tx_receipt.transactionHash,
            defaults={
                "from_address": tx_receipt["from"],
                "to_address": tx_receipt.to,
                "receipt": serialize_web3_data(tx_receipt),
            },
        )
        return tx


class AbstractTokenInfo(models.Model):
    name = models.CharField(max_length=500)
    symbol = models.CharField(max_length=16)
    decimals = models.PositiveIntegerField(default=18)

    class Meta:
        abstract = True


class NativeToken(AbstractTokenInfo):
    chain = models.OneToOneField(Chain, on_delete=models.CASCADE, related_name="native_token")


class BaseEthereumAccount(models.Model):
    address = EthereumAddressField(unique=True, db_index=True)
    transactions = models.ManyToManyField(Transaction)
    objects = InheritanceManager()

    def __str__(self):
        return self.address

    @property
    def private_key_bytes(self) -> Optional[bytes]:
        private_key = getattr(self, "private_key", None)
        return private_key and bytearray.fromhex(private_key[2:])


class Web3Provider(models.Model):
    chain = models.ForeignKey(Chain, related_name="providers", on_delete=models.CASCADE)
    url = Web3ProviderURLField()
    client_version = models.CharField(max_length=300, null=True)
    requires_geth_poa_middleware = models.BooleanField(default=False)
    supports_pending_filters = models.BooleanField(default=False)
    supports_eip1559 = models.BooleanField(default=False)
    is_active = models.BooleanField(default=True)
    synced = models.BooleanField(default=False)
    connected = models.BooleanField(default=False)

    objects = models.Manager()
    active = QueryManager(is_active=True)
    available = QueryManager(synced=True, connected=True, is_active=True)

    @property
    def is_online(self):
        return self.connected and self.synced

    @property
    def hostname(self):
        return urlparse(self.url).hostname

    @atomic()
    def activate(self):
        self.chain.providers.exclude(id=self.id).update(is_active=False)
        self.is_active = True
        self.save()

    def __str__(self):
        return self.hostname


class Explorer(models.Model):
    chain = models.ForeignKey(Chain, related_name="explorers", on_delete=models.CASCADE)
    name = models.CharField(max_length=200, null=True)
    url = models.URLField()
    standard = models.CharField(max_length=200, null=True)

    class Meta:
        unique_together = ("url", "chain")


class EventIndexer(models.Model):
    chain = models.ForeignKey(Chain, related_name="indexers", on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    last_block = models.PositiveBigIntegerField(default=1)

    @classmethod
    def make(cls, chain_id: int, name: str):
        indexer, _ = cls.objects.get_or_create(chain_id=chain_id, name=name)
        return indexer

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["chain", "name"], name="unique_name_per_chain"),
        ]


__all__ = [
    "Block",
    "Chain",
    "Transaction",
    "TransactionDataRecord",
    "BaseEthereumAccount",
    "AbstractTokenInfo",
    "NativeToken",
    "Web3Provider",
    "Explorer",
    "EventIndexer",
]



import logging

from eth_utils import to_checksum_address
from raiden_contracts.constants import (
    CONTRACT_CUSTOM_TOKEN,
    CONTRACT_SERVICE_REGISTRY,
    CONTRACT_TOKEN_NETWORK,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    CONTRACT_USER_DEPOSIT,
)
from raiden_contracts.contract_manager import (
    ContractManager,
    contracts_precompiled_path,
    get_contracts_deployment_info,
)
from raiden_contracts.utils.type_aliases import ChainID
from web3 import Web3

from hub20.apps.blockchain.client import make_web3, send_transaction
from hub20.apps.blockchain.models import Web3Provider
from hub20.apps.blockchain.typing import EthereumAccount_T
from hub20.apps.ethereum_money.abi import EIP20_ABI
from hub20.apps.ethereum_money.client import get_account_balance, make_token
from hub20.apps.ethereum_money.models import EthereumToken, EthereumTokenAmount
from hub20.apps.raiden.exceptions import InsufficientBalanceError
from hub20.apps.raiden.models import Raiden, TokenNetwork

GAS_REQUIRED_FOR_DEPOSIT: int = 200_000
GAS_REQUIRED_FOR_APPROVE: int = 70_000
GAS_REQUIRED_FOR_MINT: int = 100_000


logger = logging.getLogger(__name__)


def _get_contract_data(chain_id: int, contract_name: str):
    try:
        contract_data = get_contracts_deployment_info(ChainID(chain_id))
        assert contract_data is not None
        return contract_data["contracts"][contract_name]
    except (KeyError, AssertionError):
        return None


def get_user_deposit_contract(w3: Web3):
    contract_manager = ContractManager(contracts_precompiled_path())
    contract_address = get_contract_address(int(w3.net.version), CONTRACT_USER_DEPOSIT)
    return w3.eth.contract(
        address=contract_address, abi=contract_manager.get_contract_abi(CONTRACT_USER_DEPOSIT)
    )


def _get_contract(w3: Web3, contract_name: str):
    chain_id = int(w3.net.version)
    manager = ContractManager(contracts_precompiled_path())

    contract_data = _get_contract_data(chain_id, contract_name)
    assert contract_data

    abi = manager.get_contract_abi(contract_name)
    return w3.eth.contract(abi=abi, address=contract_data["address"])


def get_token_network_contract(w3: Web3, token_network: TokenNetwork):
    manager = ContractManager(contracts_precompiled_path())
    abi = manager.get_contract_abi(CONTRACT_TOKEN_NETWORK)
    return w3.eth.contract(abi=abi, address=token_network.address)


def get_contract_address(chain_id, contract_name):
    try:
        contract_data = _get_contract_data(chain_id, contract_name)
        return contract_data["address"]
    except (TypeError, AssertionError, KeyError) as exc:
        raise ValueError(f"{contract_name} does not exist on chain id {chain_id}") from exc


def get_token_network_registry_contract(w3: Web3):
    return _get_contract(w3, CONTRACT_TOKEN_NETWORK_REGISTRY)


def get_service_token_address(chain_id: int):
    service_contract_data = _get_contract_data(chain_id, CONTRACT_SERVICE_REGISTRY)
    return service_contract_data["constructor_arguments"][0]


def get_service_token(w3: Web3) -> EthereumToken:
    chain_id = int(w3.net.version)
    service_token_address = get_service_token_address(chain_id)
    return make_token(w3=w3, address=service_token_address)


def get_service_token_contract(w3: Web3) -> EthereumToken:
    chain_id = int(w3.net.version)
    service_token_address = get_service_token_address(chain_id)
    return w3.eth.contract(address=service_token_address, abi=EIP20_ABI)


def mint_tokens(w3: Web3, account: EthereumAccount_T, amount: EthereumTokenAmount):
    logger.debug(f"Minting {amount.formatted}")
    contract_manager = ContractManager(contracts_precompiled_path())
    token_proxy = w3.eth.contract(
        address=to_checksum_address(amount.currency.address),
        abi=contract_manager.get_contract_abi(CONTRACT_CUSTOM_TOKEN),
    )

    send_transaction(
        w3=w3,
        contract_function=token_proxy.functions.mint,
        account=account,
        contract_args=(amount.as_wei,),
        gas=GAS_REQUIRED_FOR_MINT,
    )


def make_service_deposit(w3: Web3, account: EthereumAccount_T, amount: EthereumTokenAmount):
    token = amount.currency

    on_chain_balance = get_account_balance(w3=w3, token=token, address=account.address)

    if amount > on_chain_balance:
        raise InsufficientBalanceError(
            f"Current balance of {on_chain_balance.formatted} is insufficient"
        )

    user_deposit_contract = get_user_deposit_contract(w3=w3)
    service_token_address = to_checksum_address(user_deposit_contract.functions.token().call())

    if service_token_address != token.address:
        raise ValueError(
            f"Deposit must be in {service_token_address}, {token.symbol} is {token.address}"
        )

    token_proxy = w3.eth.contract(address=token.address, abi=EIP20_ABI)
    current_deposit_amount = token.from_wei(
        user_deposit_contract.functions.total_deposit(account.address).call()
    )

    total_deposit = current_deposit_amount + amount

    old_allowance = token_proxy.functions.allowance(
        account.address, user_deposit_contract.address
    ).call()
    if old_allowance > 0:
        send_transaction(
            w3=w3,
            contract_function=token_proxy.functions.approve,
            account=account,
            contract_args=(user_deposit_contract.address, 0),
            gas=GAS_REQUIRED_FOR_APPROVE,
        )

    send_transaction(
        w3=w3,
        contract_function=token_proxy.functions.approve,
        account=account,
        contract_args=(user_deposit_contract.address, total_deposit.as_wei),
        gas=GAS_REQUIRED_FOR_APPROVE,
    )

    send_transaction(
        w3=w3,
        account=account,
        contract_function=user_deposit_contract.functions.deposit,
        contract_args=(account.address, total_deposit.as_wei),
        gas=GAS_REQUIRED_FOR_DEPOSIT,
    )


def get_service_deposit_balance(w3: Web3, raiden: Raiden) -> EthereumTokenAmount:
    user_deposit_contract = get_user_deposit_contract(w3=w3)
    token = get_service_token(w3=w3)
    return token.from_wei(user_deposit_contract.functions.effectiveBalance(raiden.address).call())


def get_token_networks():
    for provider in Web3Provider.available.exclude(raiden__isnull=True):
        w3: Web3 = make_web3(provider=provider)

        try:
            token_registry_contract = get_token_network_registry_contract(w3=w3)
        except AssertionError:
            continue
        get_token_network_address = token_registry_contract.functions.token_to_token_networks

        for token in EthereumToken.ERC20tokens.filter(chain_id=provider.chain_id):
            token_network_address = get_token_network_address(token.address).call()
            if token_network_address != EthereumToken.NULL_ADDRESS:
                TokenNetwork.objects.update_or_create(
                    token=token, defaults={"address": token_network_address}
                )



import logging

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.db.models.signals import post_save
from django.db.transaction import atomic
from django.dispatch import receiver

from hub20.apps.blockchain.models import BaseEthereumAccount, Chain, Transaction
from hub20.apps.core.models.accounting import (
    ExternalAddressAccount,
    RaidenClientAccount,
    Treasury,
    UserAccount,
    WalletAccount,
)
from hub20.apps.core.models.payments import PaymentConfirmation
from hub20.apps.core.models.transfers import (
    BlockchainTransferExecution,
    RaidenTransferExecution,
    Transfer,
    TransferCancellation,
    TransferExecution,
    TransferFailure,
)
from hub20.apps.ethereum_money.models import EthereumToken
from hub20.apps.ethereum_money.signals import incoming_transfer_mined, outgoing_transfer_mined
from hub20.apps.raiden.models import Payment as RaidenPayment, Raiden
from hub20.apps.wallet import get_wallet_model

logger = logging.getLogger(__name__)
User = get_user_model()
Wallet = get_wallet_model()


@receiver(post_save, sender=User)
def on_user_created_create_account(sender, **kw):
    if kw["created"]:
        UserAccount.objects.get_or_create(user=kw["instance"])


@receiver(post_save, sender=Chain)
def on_chain_created_create_treasury(sender, **kw):
    if kw["created"]:
        Treasury.objects.get_or_create(chain=kw["instance"])


@receiver(post_save, sender=Raiden)
def on_raiden_created_create_account(sender, **kw):
    if kw["created"]:
        RaidenClientAccount.objects.get_or_create(raiden=kw["instance"])


@receiver(post_save, sender=BaseEthereumAccount)
@receiver(post_save, sender=Raiden)
@receiver(post_save, sender=Wallet)
def on_wallet_created_create_account(sender, **kw):
    if kw["created"]:
        WalletAccount.objects.get_or_create(account=kw["instance"])


# In-Flows
@atomic()
@receiver(incoming_transfer_mined, sender=Transaction)
def on_incoming_transfer_mined_move_funds_from_external_address_to_wallet(sender, **kw):
    wallet = kw["account"]
    amount = kw["amount"]
    transaction = kw["transaction"]

    transaction_type = ContentType.objects.get_for_model(transaction)

    params = dict(
        reference_type=transaction_type,
        reference_id=transaction.id,
        currency=amount.currency,
        amount=amount.amount,
    )
    external_address_account, _ = ExternalAddressAccount.objects.get_or_create(
        address=transaction.from_address
    )

    external_address_book = external_address_account.get_book(token=amount.currency)
    wallet_book = wallet.onchain_account.get_book(token=amount.currency)

    external_address_book.debits.get_or_create(**params)
    wallet_book.credits.get_or_create(**params)


@atomic()
@receiver(post_save, sender=RaidenPayment)
def on_raiden_payment_received_move_funds_from_external_address_to_raiden(sender, **kw):
    if kw["created"]:
        payment = kw["instance"]
        raiden = payment.channel.raiden

        is_received = payment.receiver_address == raiden.address

        if is_received:
            payment_type = ContentType.objects.get_for_model(payment)
            params = dict(
                reference_type=payment_type,
                reference_id=payment.id,
                currency=payment.token,
                amount=payment.amount,
            )

            external_address_account, _ = ExternalAddressAccount.objects.get_or_create(
                address=payment.sender_address
            )

            external_account_book = external_address_account.get_book(token=payment.token)
            raiden_book = raiden.raiden_account.get_book(token=payment.token)

            external_account_book.debits.get_or_create(**params)
            raiden_book.credits.get_or_create(**params)


# Out-flows
@atomic()
@receiver(outgoing_transfer_mined, sender=Transaction)
def on_outgoing_transfer_mined_move_funds_from_wallet_to_external_address(sender, **kw):
    transaction = kw["transaction"]
    amount = kw["amount"]
    wallet = kw["account"]
    address = kw["address"]

    transaction_type = ContentType.objects.get_for_model(transaction)

    params = dict(
        reference_type=transaction_type,
        reference_id=transaction.id,
        currency=amount.currency,
        amount=amount.amount,
    )
    external_account, _ = ExternalAddressAccount.objects.get_or_create(address=address)

    wallet_book = wallet.onchain_account.get_book(token=amount.currency)
    external_account_book = external_account.get_book(token=amount.currency)

    wallet_book.debits.get_or_create(**params)
    external_account_book.credits.get_or_create(**params)


@atomic()
@receiver(post_save, sender=RaidenTransferExecution)
def on_raiden_transfer_executed_move_funds_from_raiden_to_external_address(sender, **kw):
    if kw["created"]:
        execution = kw["instance"]
        transfer = execution.transfer

        payment = execution.raidentransferexecution.payment
        transfer_type = ContentType.objects.get_for_model(transfer)
        params = dict(
            reference_type=transfer_type,
            reference_id=transfer.id,
            currency=transfer.currency,
            amount=transfer.amount,
        )

        external_account, _ = ExternalAddressAccount.objects.get_or_create(
            address=transfer.address
        )

        external_account_book = external_account.get_book(token=transfer.currency)
        raiden_book = payment.channel.raiden.raiden_account.get_book(token=transfer.currency)

        raiden_book.debits.get_or_create(**params)
        external_account_book.credits.get_or_create(**params)


@atomic()
@receiver(post_save, sender=BlockchainTransferExecution)
def on_blockchain_transfer_executed_move_fee_from_sender_to_treasury(sender, **kw):
    if kw["created"]:
        execution = kw["instance"]
        transaction = execution.transaction

        fee = execution.fee
        native_token = execution.fee.currency

        treasury_book = transaction.block.chain.treasury.get_book(token=native_token)
        sender_book = execution.transfer.sender.account.get_book(token=native_token)

        transaction_type = ContentType.objects.get_for_model(transaction)
        params = dict(
            reference_type=transaction_type,
            reference_id=transaction.id,
            currency=native_token,
            amount=fee.amount,
        )

        sender_book.debits.get_or_create(**params)
        treasury_book.credits.get_or_create(**params)


@atomic()
@receiver(post_save, sender=Transaction)
def on_transaction_submitted_move_fee_from_wallet_to_fee_account(sender, **kw):
    if kw["created"]:
        transaction = kw["instance"]

        wallet = BaseEthereumAccount.objects.filter(address=transaction.from_address).first()
        if not wallet:
            return

        native_token = EthereumToken.make_native(chain=transaction.block.chain)
        fee = native_token.from_wei(transaction.gas_fee)
        fee_account = ExternalAddressAccount.get_transaction_fee_account()

        wallet_book = wallet.onchain_account.get_book(token=native_token)
        fee_book = fee_account.get_book(token=native_token)

        transaction_type = ContentType.objects.get_for_model(transaction)
        params = dict(
            reference_type=transaction_type,
            reference_id=transaction.id,
            currency=native_token,
            amount=fee.amount,
        )

        wallet_book.debits.get_or_create(**params)
        fee_book.credits.get_or_create(**params)


# Internal movements
@atomic()
@receiver(post_save, sender=Transfer)
def on_transfer_created_move_funds_from_sender_to_treasury(sender, **kw):
    if kw["created"]:
        transfer = kw["instance"]
        params = dict(reference=transfer, currency=transfer.currency, amount=transfer.amount)

        user_book = transfer.sender.account.get_book(token=transfer.currency)
        treasury_book = transfer.currency.chain.treasury.get_book(token=transfer.currency)

        user_book.debits.create(**params)
        treasury_book.credits.create(**params)


@atomic()
@receiver(post_save, sender=TransferExecution)
def on_internal_transfer_executed_move_funds_from_treasury_to_receiver(sender, **kw):
    if kw["created"]:
        execution = kw["instance"]
        transfer = execution.transfer

        if not transfer.receiver:
            logger.warning("Expected Internal Transfer, but no receiver user defined")
            return

        params = dict(reference=transfer, currency=transfer.currency, amount=transfer.amount)

        treasury_book = transfer.currency.chain.treasury.get_book(token=transfer.currency)
        receiver_book = transfer.receiver.account.get_book(token=transfer.currency)

        treasury_book.debits.create(**params)
        receiver_book.credits.create(**params)


@atomic()
@receiver(post_save, sender=PaymentConfirmation)
def on_payment_confirmed_move_funds_from_treasury_to_payee(sender, **kw):
    if kw["created"]:
        confirmation = kw["instance"]
        payment = confirmation.payment

        is_raiden_payment = hasattr(payment.route, "raidenpaymentroute")
        is_blockchain_payment = hasattr(payment.route, "blockchainpaymentroute")

        if is_raiden_payment or is_blockchain_payment:
            params = dict(reference=confirmation, amount=payment.amount, currency=payment.currency)
            treasury_book = payment.currency.chain.treasury.get_book(token=payment.currency)
            payee_book = payment.route.deposit.user.account.get_book(token=payment.currency)
            treasury_book.debits.create(**params)
            payee_book.credits.create(**params)
        else:
            logger.info(f"Payment {payment} was not routed through any external network")


@atomic()
@receiver(post_save, sender=TransferFailure)
@receiver(post_save, sender=TransferCancellation)
def on_reverted_transaction_move_funds_from_treasury_to_sender(sender, **kw):
    if kw["created"]:
        transfer_action = kw["instance"]
        transfer = transfer_action.transfer

        if transfer.is_processed:
            logger.critical(f"{transfer} has already been processed, yet has {transfer_action}")
            return

        try:
            treasury_book = transfer.currency.chain.treasury.get_book(token=transfer.currency)
            sender_book = transfer.sender.account.get_book(token=transfer.currency)
            treasury_book.debits.create(
                reference=transfer_action, currency=transfer.currency, amount=transfer.amount
            )
            sender_book.credits.create(
                reference=transfer_action, currency=transfer.currency, amount=transfer.amount
            )

        except Exception as exc:
            logger.exception(exc)


__all__ = [
    "on_user_created_create_account",
    "on_chain_created_create_treasury",
    "on_raiden_created_create_account",
    "on_wallet_created_create_account",
    "on_incoming_transfer_mined_move_funds_from_external_address_to_wallet",
    "on_raiden_payment_received_move_funds_from_external_address_to_raiden",
    "on_outgoing_transfer_mined_move_funds_from_wallet_to_external_address",
    "on_raiden_transfer_executed_move_funds_from_raiden_to_external_address",
    "on_blockchain_transfer_executed_move_fee_from_sender_to_treasury",
    "on_transaction_submitted_move_fee_from_wallet_to_fee_account",
    "on_transfer_created_move_funds_from_sender_to_treasury",
    "on_internal_transfer_executed_move_funds_from_treasury_to_receiver",
    "on_payment_confirmed_move_funds_from_treasury_to_payee",
    "on_reverted_transaction_move_funds_from_treasury_to_sender",
]



import asyncio
import logging

from django.core.management.base import BaseCommand
from django.utils.module_loading import import_string

from hub20.apps.blockchain.client import event_streams as blockchain_streams
from hub20.apps.ethereum_money.client import event_streams as token_event_streams
from hub20.apps.raiden.client import event_streams as raiden_event_streams
from hub20.apps.raiden.client.node import sync_channels, sync_payments

from .utils import add_shutdown_handlers

logger = logging.getLogger(__name__)


BLOCKCHAIN_STREAMS = [
    blockchain_streams.process_mined_blocks,
    blockchain_streams.process_pending_transactions,
    token_event_streams.process_token_transfer_events,
    token_event_streams.process_pending_token_transfers,
    raiden_event_streams.process_channel_events,
]

INTEGRATION_STREAMS = []

RAIDEN_STREAMS = [sync_channels, sync_payments]


class Command(BaseCommand):
    help = "Runs all functions that stream events from web3 and raiden nodes"

    def add_arguments(self, parser):
        parser.add_argument(
            "--stream",
            action="extend",
            type=import_string,
            dest="streams",
            nargs="*",
        )

    def handle(self, *args, **options):
        loop = asyncio.get_event_loop()

        add_shutdown_handlers(loop)

        all_streams = BLOCKCHAIN_STREAMS + INTEGRATION_STREAMS + RAIDEN_STREAMS

        streams = options["streams"] or all_streams

        try:
            tasks = []

            for stream in streams:
                tasks.append(stream())

            asyncio.gather(*tasks, return_exceptions=True)
            loop.run_forever()
        finally:
            loop.close()



from typing import Any, Optional

from django.contrib import admin
from django.http import HttpRequest

from . import forms, models


class ReadOnlyModelAdmin(admin.ModelAdmin):
    def has_add_permission(self, request: HttpRequest, obj: Optional[Any] = None) -> bool:
        return False

    def has_change_permission(self, request: HttpRequest, obj: Optional[Any] = None) -> bool:
        return False

    def has_delete_permission(self, request: HttpRequest, obj: Optional[Any] = None) -> bool:
        return False


@admin.register(models.Raiden)
class RaidenAdmin(admin.ModelAdmin):
    form = forms.RaidenForm
    list_display = ("url", "address", "web3_provider", "chain")
    fields = (
        "url",
        "web3_provider",
        "address",
    )
    readonly_fields = ("address",)

    def has_add_permission(self, request: HttpRequest, obj: Optional[Any] = None) -> bool:
        return request.user.is_superuser


@admin.register(models.TokenNetwork)
class TokenNetworkAdmin(ReadOnlyModelAdmin):
    list_display = ("token", "address")


@admin.register(models.Payment)
class PaymentAdmin(ReadOnlyModelAdmin):
    list_display = ("channel", "token", "amount", "timestamp", "identifier", "sender_address")



import asyncio
import logging

import celery_pubsub
from asgiref.sync import sync_to_async
from raiden_contracts.constants import CONTRACT_TOKEN_NETWORK
from raiden_contracts.contract_manager import ContractManager, contracts_precompiled_path

from hub20.apps.blockchain.app_settings import BLOCK_SCAN_RANGE
from hub20.apps.blockchain.client import BLOCK_CREATION_INTERVAL, make_web3
from hub20.apps.blockchain.models import EventIndexer, Web3Provider

RAIDEN_CONTRACTS_MANAGER = ContractManager(contracts_precompiled_path())
TOKEN_NETWORK_CONTRACT_ABI = RAIDEN_CONTRACTS_MANAGER.get_contract_abi(CONTRACT_TOKEN_NETWORK)

logger = logging.getLogger(__name__)


def get_providers_on_raiden_networks():
    return Web3Provider.available.filter(chain__tokens__tokennetwork__isnull=False).select_related(
        "chain"
    )


async def process_channel_events():
    indexer_name = "raiden:token_network_channels"

    while True:
        providers = await sync_to_async(list)(get_providers_on_raiden_networks())
        for provider in providers:
            event_indexer = await sync_to_async(EventIndexer.make)(provider.chain_id, indexer_name)

            w3 = make_web3(provider=provider)
            contract = w3.eth.contract(abi=TOKEN_NETWORK_CONTRACT_ABI)

            current_block = w3.eth.block_number
            last_processed = event_indexer.last_block
            to_block = min(current_block, last_processed + BLOCK_SCAN_RANGE)

            filter_params = dict(fromBlock=last_processed, toBlock=to_block)

            for event_data in contract.events.ChannelOpened().getLogs(filter_params):
                await sync_to_async(celery_pubsub.publish)(
                    "blockchain.event.token_network_channel_opened.mined",
                    chain_id=w3.eth.chain_id,
                    event_data=event_data,
                    provider_url=provider.url,
                )

            for event_data in contract.events.ChannelClosed().getLogs(filter_params):
                await sync_to_async(celery_pubsub.publish)(
                    "blockchain.event.token_network_channel_closed.mined",
                    chain_id=w3.eth.chain_id,
                    event_data=event_data,
                    provider_url=provider.url,
                )

            event_indexer.last_block = to_block
            await sync_to_async(event_indexer.save)()
        await asyncio.sleep(BLOCK_CREATION_INTERVAL)


__all__ = ["process_channel_events"]



import random
from datetime import timedelta

import factory
from django.contrib.auth import get_user_model
from django.utils import timezone
from factory import fuzzy

from hub20.apps.blockchain.factories import BaseWalletFactory, EthereumProvider
from hub20.apps.ethereum_money.factories import Erc20TokenFactory

from . import models

factory.Faker.add_provider(EthereumProvider)


User = get_user_model()


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = User


class AdminUserFactory(factory.django.DjangoModelFactory):
    username = factory.Sequence(lambda n: f"admin-user-{n:03}")
    is_superuser = True
    is_staff = True

    class Meta:
        model = User


class TokenNetworkFactory(factory.django.DjangoModelFactory):
    address = factory.Faker("ethereum_address")
    token = factory.SubFactory(Erc20TokenFactory)

    class Meta:
        model = models.TokenNetwork


class RaidenFactory(BaseWalletFactory):
    address = "0xdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef"
    url = factory.Sequence(lambda n: "http://raiden:{0}".format(5000 + n))

    @factory.post_generation
    def channels(obj, create, extracted, **kw):
        if not create:
            return

        if not extracted:
            ChannelFactory(raiden=obj)
        else:
            for channel in extracted:
                obj.channels.add(channel)

    class Meta:
        model = models.Raiden
        django_get_or_create = ("address",)


class ChannelFactory(factory.django.DjangoModelFactory):
    raiden = factory.SubFactory(RaidenFactory)
    token_network = factory.SubFactory(TokenNetworkFactory)
    partner_address = factory.Faker("ethereum_address")
    identifier = factory.Sequence(lambda n: n)
    balance = fuzzy.FuzzyDecimal(1, 100, precision=6)
    total_deposit = factory.LazyAttribute(lambda obj: obj.balance)
    total_withdraw = 0
    status = models.Channel.STATUS.opened

    class Meta:
        model = models.Channel


class PaymentEventFactory(factory.django.DjangoModelFactory):
    channel = factory.SubFactory(ChannelFactory)
    amount = factory.fuzzy.FuzzyDecimal(0, 1, precision=4)
    timestamp = factory.LazyAttribute(
        lambda obj: PaymentEventFactory.make_sequenced_timestamp(obj)
    )
    identifier = factory.Sequence(lambda n: n)
    sender_address = factory.Faker("ethereum_address")
    receiver_address = factory.Faker("ethereum_address")

    @staticmethod
    def make_sequenced_timestamp(payment_event):
        since = payment_event.channel.last_event_timestamp or timezone.now()
        return since + timedelta(microseconds=random.randint(1, int(1e8)))

    class Meta:
        model = models.Payment


class RaidenManagementOrderFactory(factory.django.DjangoModelFactory):
    raiden = factory.SubFactory(RaidenFactory)
    user = factory.SubFactory(AdminUserFactory)


class UserDepositContractOrderFactory(RaidenManagementOrderFactory):
    currency = factory.SubFactory(Erc20TokenFactory)
    amount = factory.fuzzy.FuzzyDecimal(1, 10, precision=3)

    class Meta:
        model = models.UserDepositContractOrder


__all__ = [
    "UserFactory",
    "AdminUserFactory",
    "TokenNetworkFactory",
    "RaidenFactory",
    "ChannelFactory",
    "PaymentEventFactory",
    "UserDepositContractOrderFactory",
]



#!/usr/bin/env python

from setuptools import find_packages, setup

with open("requirements.txt") as req_file:
    install_requires = [
        req for req in req_file if req.strip() and not req.lstrip().startswith("#")
    ]

with open("README.md") as readme_file:
    description = readme_file.read()


setup(
    name="hub20",
    url="https://github.com/mushroomlabs/hub20",
    long_description=description,
    version="0.4.0",
    python_requires="~=3.7",
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    zip_safe=False,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Web Environment",
        "Framework :: Django",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
    ],
    keywords="hub20 ethereum payment-gateway",
)



from typing import Optional

from django.contrib.auth import get_user_model
from django.db.models import BooleanField, Case, ProtectedError, Value, When
from django.db.models.query import QuerySet
from django.http import Http404
from django.shortcuts import get_object_or_404
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import generics, status
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter
from rest_framework.mixins import (
    CreateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    RetrieveModelMixin,
)
from rest_framework.permissions import AllowAny, IsAdminUser, IsAuthenticated
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet, ModelViewSet

from hub20.apps.blockchain.client import make_web3
from hub20.apps.ethereum_money.client import get_estimate_fee
from hub20.apps.ethereum_money.models import EthereumToken
from hub20.apps.ethereum_money.views import BaseTokenViewSet, TokenViewSet

from . import models, serializers
from .filters import DepositFilter, UserFilter
from .permissions import IsStoreOwner

User = get_user_model()


class ReadWriteSerializerMixin(generics.GenericAPIView):
    """
    Overrides get_serializer_class to choose the read serializer
    for GET requests and the write serializer for POST requests.

    Set read_serializer_class and write_serializer_class attributes on a
    generic APIView
    """

    read_serializer_class: Optional[Serializer] = None
    write_serializer_class: Optional[Serializer] = None

    def get_serializer_class(self) -> Serializer:
        if self.request and self.request.method in ["POST", "PUT", "PATCH"]:
            return self.get_write_serializer_class()
        return self.get_read_serializer_class()

    def get_read_serializer_class(self) -> Serializer:
        assert self.read_serializer_class is not None, (
            "'%s' should either include a `read_serializer_class` attribute,"
            "or override the `get_read_serializer_class()` method." % self.__class__.__name__
        )
        return self.read_serializer_class

    def get_write_serializer_class(self) -> Serializer:
        assert self.write_serializer_class is not None, (
            "'%s' should either include a `write_serializer_class` attribute,"
            "or override the `get_write_serializer_class()` method." % self.__class__.__name__
        )
        return self.write_serializer_class


class AccountCreditEntryList(generics.ListAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.CreditSerializer

    def get_queryset(self) -> QuerySet:
        return models.Credit.objects.filter(book__account__user=self.request.user)


class AccountDebitEntryList(generics.ListAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.DebitSerializer

    def get_queryset(self) -> QuerySet:
        return self.request.user.account.debits.all()


class BaseDepositView:
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.DepositSerializer


class DepositListView(BaseDepositView, generics.ListCreateAPIView):
    filterset_class = DepositFilter
    filter_backends = (
        OrderingFilter,
        DjangoFilterBackend,
    )
    ordering = "-created"

    def get_queryset(self) -> QuerySet:
        return self.request.user.deposit_set.all()


class DepositView(BaseDepositView, generics.RetrieveAPIView):
    def get_object(self) -> models.Deposit:
        return get_object_or_404(models.Deposit, pk=self.kwargs.get("pk"), user=self.request.user)


class BasePaymentOrderView(ReadWriteSerializerMixin):
    read_serializer_class = serializers.PaymentOrderReadSerializer
    write_serializer_class = serializers.PaymentOrderSerializer


class PaymentOrderListView(BasePaymentOrderView, generics.ListCreateAPIView):
    permission_classes = (IsAuthenticated,)

    def get_queryset(self) -> QuerySet:
        return models.PaymentOrder.objects.filter(user=self.request.user)


class PaymentOrderView(BasePaymentOrderView, generics.RetrieveDestroyAPIView):
    permission_classes = (IsAuthenticated,)

    def get_object(self) -> models.PaymentOrder:
        return get_object_or_404(
            models.PaymentOrder, pk=self.kwargs.get("pk"), user=self.request.user
        )

    def destroy(self, request, pk=None):
        try:
            return super().destroy(request, pk=pk)
        except ProtectedError:
            return Response(
                "Order has either been paid or has open routes and can not be canceled",
                status=status.HTTP_400_BAD_REQUEST,
            )


class TokenBrowserViewSet(TokenViewSet):
    def get_serializer_class(self):
        if self.action == "balance":
            return serializers.HyperlinkedTokenBalanceSerializer

        return super().get_serializer_class()

    @action(detail=True)
    def transfer_cost(self, request, **kwargs):
        """
        Returns estimated cost in Wei (estimated gas * gas price) to execute a transfer

        Returns 404 if not connected to the blockchain or if token not in database
        """
        token = self.get_object()
        try:
            w3 = make_web3(provider=token.chain.provider)
            transfer_cost = get_estimate_fee(w3=w3, token=token)
            return Response(transfer_cost.as_wei)
        except AttributeError:
            raise Http404
        except TypeError:
            return Response(status=status.HTTP_503_SERVICE_UNAVAILABLE)

    @action(detail=True, permission_classes=(IsAuthenticated,))
    def balance(self, request, **kwargs):
        """
        Returns user balance for that token
        """
        try:
            token = self.get_object()
            balance = self.request.user.account.get_balance(token)
            serializer = self.get_serializer(instance=balance)
            return Response(serializer.data)
        except AttributeError:
            raise Http404


class TransferListView(generics.ListCreateAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.TransferSerializer

    def get_queryset(self) -> QuerySet:
        return self.request.user.transfers_sent.all()


class TransferView(generics.RetrieveAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.TransferSerializer

    def get_object(self):
        try:
            return models.Transfer.objects.get(pk=self.kwargs.get("pk"), sender=self.request.user)
        except models.Transfer.DoesNotExist:
            raise Http404


class TokenBalanceListView(generics.ListAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.HyperlinkedTokenBalanceSerializer
    filter_backends = (OrderingFilter,)
    ordering = ("chain_id", "-is_native", "symbol")

    def get_queryset(self) -> QuerySet:
        return self.request.user.account.get_balances().annotate(
            is_native=Case(
                When(address=EthereumToken.NULL_ADDRESS, then=Value(True)),
                default=Value(False),
                output_field=BooleanField(),
            )
        )


class TokenBalanceView(generics.RetrieveAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.HyperlinkedTokenBalanceSerializer

    def get_object(self) -> EthereumToken:
        token = get_object_or_404(
            EthereumToken, chain_id=self.kwargs["chain_id"], address=self.kwargs["address"]
        )
        return self.request.user.account.get_balance(token)


class CheckoutViewSet(GenericViewSet, CreateModelMixin, RetrieveModelMixin):
    permission_classes = (AllowAny,)
    serializer_class = serializers.HttpCheckoutSerializer
    lookup_value_regex = "[0-9a-f-]{36}"

    def get_queryset(self):
        return models.Checkout.objects.all()

    def get_object(self):
        return get_object_or_404(models.Checkout, id=self.kwargs["pk"])


class PaymentViewSet(GenericViewSet, ListModelMixin, RetrieveModelMixin):
    lookup_value_regex = "[0-9a-f-]{36}"

    def get_queryset(self):
        return models.Payment.objects.all()

    def get_permissions(self):
        permission_class = IsAdminUser if self.action == "list" else AllowAny
        return (permission_class(),)

    def get_serializer_class(self):
        if self.action == "list":
            return serializers.PaymentSerializer

        payment = self.get_object()

        return {
            models.InternalPayment: serializers.InternalPaymentSerializer,
            models.BlockchainPayment: serializers.BlockchainPaymentSerializer,
            models.RaidenPayment: serializers.RaidenPaymentSerializer,
        }.get(type(payment), serializers.PaymentSerializer)

    def get_object(self):
        try:
            return models.Payment.objects.get_subclass(id=self.kwargs["pk"])
        except (models.Payment.DoesNotExist, KeyError):
            return None


class StoreViewSet(GenericViewSet, ListModelMixin, RetrieveModelMixin):
    permission_classes = (AllowAny,)
    serializer_class = serializers.StoreViewerSerializer
    queryset = models.Store.objects.all()

    def get_object(self, *args, **kw):
        return get_object_or_404(models.Store, id=self.kwargs["pk"])


class UserStoreViewSet(ModelViewSet):
    permission_classes = (IsStoreOwner,)
    serializer_class = serializers.StoreEditorSerializer

    def get_queryset(self) -> QuerySet:
        try:
            return self.request.user.store_set.all()
        except AttributeError:
            return models.Store.objects.none()

    def get_object(self, *args, **kw):
        store = get_object_or_404(models.Store, id=self.kwargs["pk"])
        self.check_object_permissions(self.request, store)
        return store


class UserTokenViewSet(BaseTokenViewSet, DestroyModelMixin):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.UserTokenSerializer

    def get_serializer_class(self):
        if self.action == "create":
            return serializers.UserTokenCreatorSerializer

        return self.serializer_class

    def get_queryset(self) -> QuerySet:
        qs = super().get_queryset()
        return qs.filter(userpreferences__user=self.request.user)

    def destroy(self, *args, **kw):
        token = self.get_object()
        self.request.user.preferences.tokens.remove(token)
        return Response(status.HTTP_204_NO_CONTENT)


class UserPreferencesView(generics.RetrieveUpdateAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.UserPreferencesSerializer

    def get_object(self) -> QuerySet:
        return self.request.user.preferences


class UserViewSet(GenericViewSet, ListModelMixin, RetrieveModelMixin):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.UserSerializer
    filterset_class = UserFilter
    filter_backends = (
        OrderingFilter,
        DjangoFilterBackend,
    )
    lookup_field = "username"
    ordering = "username"

    def get_queryset(self) -> QuerySet:
        return User.objects.filter(is_active=True, is_superuser=False, is_staff=False)

    def get_object(self, *args, **kw):
        return get_object_or_404(
            User,
            is_active=True,
            is_superuser=False,
            is_staff=False,
            username=self.kwargs["username"],
        )


class StatusView(APIView):
    permission_classes = (IsAuthenticated,)


class AccountingReportView(StatusView):
    permission_classes = (IsAdminUser,)

    def _get_serialized_book(self, accounting_model_class):
        books = accounting_model_class.balance_sheet().exclude(total_credit=0, total_debit=0)
        return serializers.AccountingBookSerializer(
            books, many=True, context={"request": self.request}
        ).data

    def get(self, request, **kw):
        return Response(
            dict(
                treasury=self._get_serialized_book(models.Treasury),
                wallets=self._get_serialized_book(models.WalletAccount),
                raiden=self._get_serialized_book(models.RaidenClientAccount),
                user_accounts=self._get_serialized_book(models.UserAccount),
                external_addresses=self._get_serialized_book(models.ExternalAddressAccount),
            )
        )


class BalanceSheetWalletViewSet(GenericViewSet, ListModelMixin, RetrieveModelMixin):
    permission_classes = (IsAdminUser,)
    serializer_class = serializers.WalletBalanceSheetSerializer
    lookup_url_kwarg = "address"
    lookup_field = "account__address"

    def get_queryset(self) -> QuerySet:
        return models.WalletAccount.objects.all()

    def get_object(self) -> models.WalletAccount:
        return get_object_or_404(models.WalletAccount, account__address=self.kwargs.get("address"))



import factory

from hub20.apps.core import models
from hub20.apps.ethereum_money.factories import UserTokenListFactory

from .base import UserFactory
from .payments import Erc20TokenPaymentOrderFactory


class StoreFactory(factory.django.DjangoModelFactory):
    owner = factory.SubFactory(UserFactory)
    name = factory.Sequence(lambda n: f"Store #{n:02}")
    url = factory.Sequence(lambda n: f"https://store#{n:02}.example.com")
    accepted_token_list = factory.SubFactory(
        UserTokenListFactory, user=factory.SelfAttribute("..owner")
    )
    checkout_webhook_url = factory.Sequence(lambda n: f"https://store#{n:02}.example.com/checkout")

    class Meta:
        model = models.Store


class CheckoutFactory(Erc20TokenPaymentOrderFactory):
    store = factory.SubFactory(StoreFactory, accepted_token_list__tokens=[])
    external_identifier = factory.fuzzy.FuzzyText(length=30, prefix="checkout-")
    user = factory.SelfAttribute(".store.owner")

    class Meta:
        model = models.Checkout


__all__ = ["StoreFactory", "CheckoutFactory"]



from typing import TypeVar, Union

from hexbytes import HexBytes

from .fields import EthereumAddressField

Address = Union[str, HexBytes, EthereumAddressField]
EthereumAccount_T = TypeVar("EthereumAccount_T")
TransactionHash = HexBytes



from __future__ import annotations

import json
import logging
import uuid

import requests
from django.contrib.auth import get_user_model
from django.db import models
from django.db.models import Q, Sum
from model_utils.managers import QueryManager
from model_utils.models import TimeStampedModel
from taggit.managers import TaggableManager

from hub20.apps.blockchain.fields import EthereumAddressField
from hub20.apps.blockchain.models import Chain

from . import choices
from .fields import EthereumTokenAmountField, TokenlistStandardURLField
from .schemas import TokenList as TokenListDataModel, validate_token_list
from .typing import TokenAmount, TokenAmount_T, Wei

logger = logging.getLogger(__name__)
User = get_user_model()


class EthereumToken(models.Model):
    NULL_ADDRESS = "0x0000000000000000000000000000000000000000"
    chain = models.ForeignKey(Chain, on_delete=models.CASCADE, related_name="tokens")
    address = EthereumAddressField(default=NULL_ADDRESS)
    symbol = models.CharField(max_length=20)
    name = models.CharField(max_length=500)
    decimals = models.PositiveIntegerField(default=18)
    logoURI = TokenlistStandardURLField(max_length=512, null=True, blank=True)
    is_listed = models.BooleanField(default=False)

    objects = models.Manager()
    native = QueryManager(address=NULL_ADDRESS)
    ERC20tokens = QueryManager(~Q(address=NULL_ADDRESS))
    tradeable = QueryManager(chain__providers__is_active=True)

    @property
    def is_ERC20(self) -> bool:
        return self.address != self.NULL_ADDRESS

    @property
    def wrapped_by(self):
        return self.__class__.objects.filter(id__in=self.wrapping_tokens.values("wrapper"))

    @property
    def wraps(self):
        wrapping = getattr(self, "wrappedtoken", None)
        return wrapping and wrapping.wrapper

    @property
    def is_stable(self):
        return hasattr(self, "stable_pair")

    @property
    def tracks_currency(self):
        pairing = getattr(self, "stable_pair", None)
        return pairing and pairing.currency

    def __str__(self) -> str:
        components = [self.symbol]
        if self.is_ERC20:
            components.append(self.address)

        components.append(str(self.chain_id))
        return " - ".join(components)

    def from_wei(self, wei_amount: Wei) -> EthereumTokenAmount:
        value = TokenAmount(wei_amount) / (10 ** self.decimals)
        return EthereumTokenAmount(amount=value, currency=self)

    @classmethod
    def make_native(cls, chain: Chain):
        token, _ = cls.objects.update_or_create(
            chain=chain,
            address=cls.NULL_ADDRESS,
            defaults=dict(
                is_listed=True,
                name=chain.native_token.name,
                decimals=chain.native_token.decimals,
                symbol=chain.native_token.symbol,
            ),
        )
        return token

    @classmethod
    def make(cls, address: str, chain: Chain, **defaults):
        if address == cls.NULL_ADDRESS:
            return cls.make_native(chain)

        obj, _ = cls.objects.update_or_create(address=address, chain=chain, defaults=defaults)
        return obj

    class Meta:
        unique_together = (("chain", "address"),)


class WrappedToken(models.Model):
    """
    Tokens that have are minted and burned dependent on how much
    of a primary token is locked are called 'wrapper tokens'. This can
    happen in-chain (e.g, W-ETH wrapping ETH), or cross-chain
    "bridged" tokens (e.g, WBTC wrapping BTC, Binance BAT wrapping
    ERC20 BAT, Arbitrum DAI wrapping Ethereum DAI, etc)a

    The idea of this table is to provide a way for operators and users
    to optionally accept payments and transfers of a wrapped token
    instead of the 'original' one.

    """

    wrapped = models.ForeignKey(
        EthereumToken, related_name="wrapping_tokens", on_delete=models.CASCADE
    )
    wrapper = models.OneToOneField(EthereumToken, on_delete=models.CASCADE)

    class Meta:
        unique_together = ("wrapped", "wrapper")


class StableTokenPair(models.Model):
    """
    A stabletoken is a token whose value is supposed to be pegged to a
    'real' fiat currency. These value pegs can be 'soft' or 'hard',
    (e.g, both USDC and DAI are pegged to the USD, but DAI's price is
    determined algorithmically and therefore can oscillate, while USDC
    is supposed to always be worth one USD
    """

    token = models.OneToOneField(
        EthereumToken, related_name="stable_pair", on_delete=models.CASCADE
    )
    algorithmic_peg = models.BooleanField(default=True)
    currency = models.CharField(max_length=3, choices=choices.CURRENCIES)


class AbstractTokenListModel(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    name = models.CharField(max_length=64)
    description = models.TextField(null=True)
    tokens = models.ManyToManyField(
        EthereumToken, related_name="%(app_label)s_%(class)s_tokenlists"
    )
    keywords = TaggableManager()

    def __str__(self):
        return self.name

    @classmethod
    def fetch(cls, url) -> TokenListDataModel:
        response = requests.get(url)

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            raise ValueError(f"Failed to fetch {url}")

        try:
            token_list_data = response.json()
        except json.decoder.JSONDecodeError:
            raise ValueError(f"Could not decode json response from {url}")

        validate_token_list(token_list_data)

        return TokenListDataModel(**token_list_data)

    class Meta:
        abstract = True


class TokenList(AbstractTokenListModel):
    """
    A model to manage [token lists](https://tokenlists.org). Only
    admins can manage/import/export them.
    """

    url = TokenlistStandardURLField()
    version = models.CharField(max_length=32)

    class Meta:
        unique_together = ("url", "version")

    @classmethod
    def make(cls, url, token_list_data: TokenListDataModel, description=None):

        token_list, _ = cls.objects.get_or_create(
            url=url,
            version=token_list_data.version.as_string,
            defaults=dict(name=token_list_data.name),
        )
        token_list.description = description
        token_list.keywords.add(*token_list_data.keywords)
        token_list.save()

        for token_entry in token_list_data.tokens:
            token, _ = EthereumToken.objects.get_or_create(
                chain_id=token_entry.chainId,
                address=token_entry.address,
                defaults=dict(
                    name=token_entry.name,
                    decimals=token_entry.decimals,
                    symbol=token_entry.symbol,
                    logoURI=token_entry.logoURI,
                ),
            )
            token_list.tokens.add(token)
        return token_list


class UserTokenList(AbstractTokenListModel, TimeStampedModel):
    """
    A model to manage [token lists](https://tokenlists.org). Only
    admins can manage/import/export them.
    """

    user = models.ForeignKey(User, related_name="token_lists", on_delete=models.CASCADE)

    @classmethod
    def clone(cls, user, token_list: TokenList):
        user_token_list = user.token_lists.create(
            name=token_list.name,
            description=token_list.description,
        )
        user_token_list.tokens.set(token_list.tokens.all())
        user_token_list.keywords.set(token_list.keywords.all())
        return user_token_list


class EthereumTokenValueModel(models.Model):
    amount = EthereumTokenAmountField()
    currency = models.ForeignKey(EthereumToken, on_delete=models.PROTECT)

    @property
    def as_token_amount(self):
        return EthereumTokenAmount(amount=self.amount, currency=self.currency)

    @property
    def formatted_amount(self):
        return self.as_token_amount.formatted

    class Meta:
        abstract = True


class EthereumTokenAmount:
    def __init__(self, amount: TokenAmount_T, currency: EthereumToken):
        self.amount: TokenAmount = TokenAmount(amount)
        self.currency: EthereumToken = currency

    @property
    def formatted(self):
        integral = int(self.amount)
        frac = self.amount % 1

        amount_formatted = str(integral) if not bool(frac) else self.amount.normalize()
        return f"{amount_formatted} {self.currency.symbol}"

    @property
    def as_wei(self) -> Wei:
        return Wei(self.amount * (10 ** self.currency.decimals))

    @property
    def as_hex(self) -> str:
        return hex(self.as_wei)

    @property
    def is_native_token(self) -> bool:
        return self.currency.address == EthereumToken.NULL_ADDRESS

    def _check_currency_type(self, other: EthereumTokenAmount):
        if not self.currency == other.currency:
            raise ValueError(f"Can not operate {self.currency} and {other.currency}")

    def __add__(self, other: EthereumTokenAmount) -> EthereumTokenAmount:
        self._check_currency_type(self)
        return self.__class__(self.amount + other.amount, self.currency)

    def __sub__(self, other: EthereumTokenAmount) -> EthereumTokenAmount:
        self._check_currency_type(self)
        return self.__class__(self.amount - other.amount, self.currency)

    def __mul__(self, other: TokenAmount_T) -> EthereumTokenAmount:
        return EthereumTokenAmount(amount=TokenAmount(other) * self.amount, currency=self.currency)

    def __rmul__(self, other: TokenAmount_T) -> EthereumTokenAmount:
        return self.__mul__(other)

    def __eq__(self, other: object) -> bool:
        message = f"Can not compare {self.currency} amount with {type(other)}"
        assert isinstance(other, EthereumTokenAmount), message

        return self.currency == other.currency and self.amount == other.amount

    def __lt__(self, other: EthereumTokenAmount):
        self._check_currency_type(other)
        return self.amount < other.amount

    def __le__(self, other: EthereumTokenAmount):
        self._check_currency_type(other)
        return self.amount <= other.amount

    def __gt__(self, other: EthereumTokenAmount):
        self._check_currency_type(other)
        return self.amount > other.amount

    def __ge__(self, other: EthereumTokenAmount):
        self._check_currency_type(other)
        return self.amount >= other.amount

    def __str__(self):
        return self.formatted

    def __repr__(self):
        return self.formatted

    @classmethod
    def aggregated(cls, queryset, currency: EthereumToken):
        entries = queryset.filter(currency=currency)
        amount = entries.aggregate(total=Sum("amount")).get("total") or TokenAmount(0)
        return cls(amount=amount, currency=currency)


__all__ = [
    "EthereumToken",
    "EthereumTokenAmount",
    "EthereumTokenValueModel",
    "StableTokenPair",
    "TokenList",
    "UserTokenList",
    "WrappedToken",
]



from django.dispatch import Signal

block_sealed = Signal(["chain_id", "block_data"])
chain_status_synced = Signal(["chain_id", "current_block", "synced"])
chain_reorganization_detected = Signal(["chain_id", "new_block_height"])
transaction_broadcast = Signal(["chain_id", "transaction_data"])



import logging
from typing import Optional

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import F, OuterRef, Subquery, Sum
from django.db.models.functions import Coalesce
from django.db.models.query import QuerySet
from model_utils.models import TimeStampedModel

from hub20.apps.blockchain.fields import EthereumAddressField
from hub20.apps.blockchain.models import BaseEthereumAccount, Chain
from hub20.apps.ethereum_money.models import (
    EthereumToken,
    EthereumTokenAmount,
    EthereumTokenAmountField,
    EthereumTokenValueModel,
)
from hub20.apps.raiden.models import Raiden

logger = logging.getLogger(__name__)


class DoubleEntryAccountModelQuerySet(models.QuerySet):
    def grouped_by_token_balances(self):
        credit = Coalesce(Sum("books__credits__amount"), 0, output_field=models.DecimalField())
        debit = Coalesce(Sum("books__debits__amount"), 0, output_field=models.DecimalField())

        return (
            self.exclude(books__token=None)
            .annotate(token_id=F("books__token"))
            .annotate(total_credit=credit, total_debit=debit)
            .annotate(balance=F("total_credit") - F("total_debit"))
        )

    def with_funds(self, token_amount: EthereumTokenAmount):
        return self.grouped_by_token_balances().filter(
            balance__gte=token_amount.amount, token_id=token_amount.currency.id
        )


class Book(models.Model):
    token = models.ForeignKey(EthereumToken, on_delete=models.PROTECT, related_name="books")
    owner_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    owner_id = models.PositiveIntegerField()
    owner = GenericForeignKey("owner_type", "owner_id")

    class Meta:
        unique_together = ("token", "owner_type", "owner_id")


class BookEntry(TimeStampedModel, EthereumTokenValueModel):
    book = models.ForeignKey(Book, on_delete=models.CASCADE, related_name="entries")
    reference_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    reference_id = models.PositiveIntegerField()
    reference = GenericForeignKey("reference_type", "reference_id")

    def clean(self):
        if self.book.token != self.currency:
            raise ValidationError(
                f"Can not add a {self.currency} entry to a {self.book.token} book"
            )

    class Meta:
        abstract = True
        unique_together = ("book", "reference_type", "reference_id")


class Credit(BookEntry):
    book = models.ForeignKey(Book, on_delete=models.CASCADE, related_name="credits")


class Debit(BookEntry):
    book = models.ForeignKey(Book, on_delete=models.CASCADE, related_name="debits")


class DoubleEntryAccountModel(models.Model):
    book_relation_attr = None
    token_balance_relation_attr = None
    objects = DoubleEntryAccountModelQuerySet.as_manager()

    @property
    def debits(self):
        return Debit.objects.filter(**{self.book_relation_attr: self})

    @property
    def credits(self):
        return Credit.objects.filter(**{self.book_relation_attr: self})

    @property
    def balances(self):
        return self.get_balances()

    def get_book(self, token: EthereumToken) -> Book:
        book, _ = self.books.get_or_create(token=token)
        return book

    def get_balance_token_amount(self, token: EthereumToken) -> Optional[EthereumTokenAmount]:
        balance = self.get_balance(token=token)
        return balance and EthereumTokenAmount(currency=token, amount=balance.balance)

    def get_balance(self, token: EthereumToken) -> Optional[EthereumToken]:
        return self.get_balances().filter(id=token.id).first()

    def get_balances(self) -> QuerySet:
        total_sum = Coalesce(Sum("amount"), 0, output_field=EthereumTokenAmountField())
        credit_qs = self.credits.values(token=F("book__token")).annotate(total_credit=total_sum)
        debit_qs = self.debits.values(token=F("book__token")).annotate(total_debit=total_sum)

        credit_sqs = credit_qs.filter(token=OuterRef("pk"))
        debit_sqs = debit_qs.filter(token=OuterRef("pk"))

        annotated_qs = EthereumToken.objects.annotate(
            total_credit=Coalesce(
                Subquery(credit_sqs.values("total_credit")),
                0,
                output_field=EthereumTokenAmountField(),
            ),
            total_debit=Coalesce(
                Subquery(debit_sqs.values("total_debit")),
                0,
                output_field=EthereumTokenAmountField(),
            ),
        )
        return annotated_qs.annotate(balance=F("total_credit") - F("total_debit")).exclude(
            total_credit=0, total_debit=0
        )

    @classmethod
    def balance_sheet(cls):
        total_sum = Coalesce(Sum("amount"), 0, output_field=EthereumTokenAmountField())
        filter_q = {f"{cls.book_relation_attr}__isnull": False}
        credit_qs = (
            Credit.objects.filter(**filter_q)
            .values(token=F("book__token"))
            .annotate(total_credit=total_sum)
        )
        debit_qs = (
            Debit.objects.filter(**filter_q)
            .values(token=F("book__token"))
            .annotate(total_debit=total_sum)
        )

        credit_sqs = credit_qs.filter(token=OuterRef("pk"))
        debit_sqs = debit_qs.filter(token=OuterRef("pk"))

        annotated_qs = EthereumToken.objects.annotate(
            total_credit=Coalesce(
                Subquery(credit_sqs.values("total_credit")),
                0,
                output_field=EthereumTokenAmountField(),
            ),
            total_debit=Coalesce(
                Subquery(debit_sqs.values("total_debit")),
                0,
                output_field=EthereumTokenAmountField(),
            ),
        )
        return annotated_qs.annotate(balance=F("total_credit") - F("total_debit"))

    class Meta:
        abstract = True


##############################################################################
#
# The following diagram illustrates how the different accounts part of the
# system have their funds accounted for.
#
#
#      Ethereum Accounts               Raiden Client
#              +                           +
#              |                           |
#              |       +----------+        |
#              +------>+          +<-------+
#                      | Treasury |
#              +------>+          +<-------+
#              |       +----+-----+        |
#              |                           |
#              +                           |
#   User Balances                     External Addresses
#
#
# The funds from Ethereum Accounts and Raiden Clients are managed by the
# treasury. Every blockchain transaction or payment received through the raiden
# network counts as a credit for the treasury and a debit from the originating
# external address (which can be regular accounts or smart contracts). The
# Treasury forwards along any credit received to the wallet that was the target
# of the transaction, and in this way we can also keep the balance of each
# wallet without having to query the blockchain every time.
#
# User funds are accounted in a similar manner. Transfers done by the user
# should be treated as a credit to the treasury, and payments related to
# payment orders should lead to a credit to the user.
#
# We should at the very least have the following equations being satisfied per
# token, else the site should be considered insolvent:
#
#        (I)   Assets = Treasury + Ethereum Accounts + Raiden
#       (II)   Assets >= User Balances
#
# All of these operations are now defined at the handlers.accounting module.
#
#############################################################################
class Treasury(DoubleEntryAccountModel):
    book_relation_attr = "book__treasury"
    token_balance_relation_attr = "books__treasury"

    chain = models.OneToOneField(Chain, on_delete=models.CASCADE)
    books = GenericRelation(
        Book,
        content_type_field="owner_type",
        object_id_field="owner_id",
        related_query_name="treasury",
    )


class UserAccount(DoubleEntryAccountModel):
    book_relation_attr = "book__account"
    token_balance_relation_attr = "books__account"

    user = models.OneToOneField(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="account"
    )
    books = GenericRelation(
        Book,
        content_type_field="owner_type",
        object_id_field="owner_id",
        related_query_name="account",
    )


class RaidenClientAccount(DoubleEntryAccountModel):
    book_relation_attr = "book__raiden"
    token_balance_relation_attr = "books__raiden"

    raiden = models.OneToOneField(Raiden, on_delete=models.CASCADE, related_name="raiden_account")
    books = GenericRelation(
        Book,
        content_type_field="owner_type",
        object_id_field="owner_id",
        related_query_name="raiden",
    )


class WalletAccount(DoubleEntryAccountModel):
    book_relation_attr = "book__wallet"
    token_balance_relation_attr = "books__wallet"

    account = models.OneToOneField(
        BaseEthereumAccount, on_delete=models.CASCADE, related_name="onchain_account"
    )
    books = GenericRelation(
        Book,
        content_type_field="owner_type",
        object_id_field="owner_id",
        related_query_name="wallet",
    )


class ExternalAddressAccount(DoubleEntryAccountModel):
    book_relation_attr = "book__address"
    token_balance_relation_attr = "books__address"

    address = EthereumAddressField(unique=True)
    books = GenericRelation(
        Book,
        content_type_field="owner_type",
        object_id_field="owner_id",
        related_query_name="address",
    )

    @classmethod
    def get_transaction_fee_account(cls):
        account, _ = cls.objects.get_or_create(address=EthereumToken.NULL_ADDRESS)
        return account


__all__ = [
    "Book",
    "BookEntry",
    "Credit",
    "Debit",
    "Treasury",
    "RaidenClientAccount",
    "WalletAccount",
    "UserAccount",
    "ExternalAddressAccount",
]



from model_utils.choices import Choices


ETHEREUM_CHAINS = Choices(
    (1, "mainnet", "Mainnet"),
    (2, "testnet", "Test Network"),
    (3, "ropsten", "Ropsten"),
    (4, "rinkeby", "Rinkeby"),
    (5, "goerli", "Görli"),
    (42, "kovan", "Kovan"),
)



from django.urls import path
from rest_framework.routers import SimpleRouter

from hub20.apps.ethereum_money.api import TokenRouter
from hub20.apps.ethereum_money.views import TokenListViewSet, UserTokenListViewSet

from . import views
from .consumers import CheckoutConsumer, SessionEventsConsumer

app_name = "hub20"

token_router = TokenRouter(trailing_slash=False)
token_router.register(r"tokens", views.TokenBrowserViewSet, basename="token")
token_router.register(r"my/tokens", views.UserTokenViewSet, basename="user-token")


router = SimpleRouter(trailing_slash=False)
router.register("checkout", views.CheckoutViewSet, basename="checkout")
router.register("payments", views.PaymentViewSet, basename="payments")
router.register("tokenlists", TokenListViewSet, basename="tokenlist")
router.register("stores", views.StoreViewSet, basename="store")
router.register("users", views.UserViewSet, basename="users")
router.register(
    "accounting/wallets", views.BalanceSheetWalletViewSet, basename="accounting-wallets"
)
router.register("my/stores", views.UserStoreViewSet, basename="user-store")
router.register("my/tokenlists", UserTokenListViewSet, basename="user-tokenlist")

urlpatterns = (
    [
        path("credits", views.AccountCreditEntryList.as_view(), name="credit-list"),
        path("debits", views.AccountDebitEntryList.as_view(), name="debit-list"),
        path("balances", views.TokenBalanceListView.as_view(), name="balance-list"),
        path(
            "balances/<int:chain_id>-<str:address>",
            views.TokenBalanceView.as_view(),
            name="balance-detail",
        ),
        path("deposits", views.DepositListView.as_view(), name="deposit-list"),
        path("deposit/<uuid:pk>", views.DepositView.as_view(), name="deposit-detail"),
        path("payment/orders", views.PaymentOrderListView.as_view(), name="payment-order-list"),
        path(
            "payment/order/<uuid:pk>",
            views.PaymentOrderView.as_view(),
            name="payment-order-detail",
        ),
        path("transfers", views.TransferListView.as_view(), name="transfer-list"),
        path("transfers/transfer/<int:pk>", views.TransferView.as_view(), name="transfer-detail"),
        path("accounting/report", views.AccountingReportView.as_view(), name="accounting-report"),
        path("my/settings", views.UserPreferencesView.as_view(), name="user-preferences"),
    ]
    + router.urls
    + token_router.urls
)


consumer_patterns = [
    path("events", SessionEventsConsumer.as_asgi()),
    path("checkout/<uuid:pk>", CheckoutConsumer.as_asgi()),
]



import datetime
import logging
import random
import uuid
from typing import Optional

from django.conf import settings
from django.contrib.postgres.fields.ranges import IntegerRangeField
from django.db import models
from django.db.models import Exists, ExpressionWrapper, F, OuterRef, Q, Sum, Value
from django.db.models.functions import Coalesce, Lower, Upper
from django.utils import timezone
from model_utils.managers import InheritanceManager
from model_utils.models import TimeStampedModel

from hub20.apps.blockchain.models import BaseEthereumAccount, Chain, Transaction
from hub20.apps.ethereum_money.models import (
    EthereumToken,
    EthereumTokenAmountField,
    EthereumTokenValueModel,
)
from hub20.apps.raiden.models import Payment as RaidenPaymentEvent, Raiden

from ..choices import DEPOSIT_STATUS
from ..settings import app_settings

logger = logging.getLogger(__name__)


def generate_payment_order_id():
    # Default payment identifier generated by Raiden's web UI is based on unix
    # time. We would like to make the distinction between payment ids
    # generated by default and those generated by us.

    LOWER_BOUND = 2 ** 48  # Enough to take us to the year 10889.
    UPPER_BOUND = 2 ** 53 - 1  # Javascript can not handle numbers bigger than 2^53 - 1

    return random.randint(LOWER_BOUND, UPPER_BOUND)


def calculate_raiden_payment_window():
    return datetime.timedelta(seconds=app_settings.Payment.raiden_route_lifetime)


class DepositQuerySet(models.QuerySet):
    def expired(self, block_number: Optional[int] = None, at: Optional[datetime.datetime] = None):
        return self.without_blockchain_route(block_number=block_number).without_raiden_route(at=at)

    def open(self, block_number: Optional[int] = None, at: Optional[datetime.datetime] = None):
        exists_blockchain_route = self.__class__.get_blockchain_window_query(
            block_number=block_number
        )
        exists_raiden_route = self.__class__.get_raiden_window_query(at=at)

        return self.filter(exists_blockchain_route | exists_raiden_route)

    def with_blockchain_route(self, block_number: Optional[int] = None):
        exists_route = self.__class__.get_blockchain_window_query(block_number=block_number)
        return self.filter(exists_route)

    def with_raiden_route(self, at: Optional[datetime.datetime] = None):
        exists_route = self.__class__.get_raiden_window_query(at=at)
        return self.filter(exists_route)

    def without_blockchain_route(self, block_number: Optional[int] = None):
        exists_route = self.__class__.get_blockchain_window_query(block_number=block_number)
        return self.filter(~exists_route)

    def without_raiden_route(self, at: Optional[datetime.datetime] = None):
        exists_route = self.__class__.get_raiden_window_query(at=at)
        return self.filter(~exists_route)

    @classmethod
    def get_blockchain_window_query(cls, block_number: Optional[int] = None) -> Exists:
        qs = BlockchainPaymentRoute.objects.available(block_number=block_number)
        return Exists(qs.filter(deposit=OuterRef("pk")))

    @classmethod
    def get_raiden_window_query(cls, at: Optional[datetime.datetime] = None) -> Exists:
        return Exists(RaidenPaymentRoute.objects.available(at=at).filter(deposit=OuterRef("pk")))


class PaymentOrderQuerySet(DepositQuerySet):
    def unpaid(self):
        q_no_payment = Q(total_paid__isnull=True)
        q_low_payment = Q(total_paid__lt=F("amount"))

        return self.annotate(total_paid=Sum("routes__payment__amount")).filter(
            q_no_payment | q_low_payment
        )

    def paid(self):
        return self.annotate(total_paid=Sum("routes__payment__amount")).filter(
            total_paid__gte=F("amount")
        )


class BlockchainRouteQuerySet(models.QuerySet):
    def with_provider(self) -> models.QuerySet:
        return self.filter(chain__providers__is_active=True)

    def with_expiration(self) -> models.QuerySet:
        return self.annotate(
            start_block=Lower("payment_window"), expiration_block=Upper("payment_window")
        )

    def with_payment_amounts(self) -> models.QuerySet:
        return self.annotate(
            currency=F("payment__currency"),
            total_paid=Coalesce(
                Sum("payment__amount"), Value(0), output_field=EthereumTokenAmountField()
            ),
            total_confirmed=Coalesce(
                Sum("payment__amount", filter=Q(payment__confirmation__isnull=False)),
                Value(0),
                output_field=EthereumTokenAmountField(),
            ),
        )

    def expired(self, block_number: Optional[int] = None) -> models.QuerySet:
        at_block = block_number if block_number is not None else F("chain__highest_block")
        return self.filter(expiration_block__lt=at_block)

    def used(self) -> models.QuerySet:
        return self.with_payment_amounts().filter(
            total_paid__gte=F("deposit__amount"), currency=F("deposit__currency")
        )

    def available(self, block_number: Optional[int] = None) -> models.QuerySet:
        qs = self.with_expiration()
        at_block = block_number if block_number is not None else F("chain__highest_block")

        return qs.filter(start_block__lte=at_block, expiration_block__gte=at_block)

    def open(self, block_number: Optional[int] = None) -> models.QuerySet:
        at_block = block_number if block_number is not None else F("chain__highest_block")

        no_defined_amount = Q(deposit__paymentorder__isnull=True)

        confirmed = Q(total_confirmed__gte=F("deposit__paymentorder__amount")) & Q(
            currency=F("deposit__currency")
        )
        expired = Q(expiration_block__lt=at_block)

        return (
            self.with_expiration()
            .exclude(expired)
            .with_payment_amounts()
            .filter(no_defined_amount | ~confirmed)
        )


class RaidenRouteQuerySet(models.QuerySet):
    def with_expiration(self) -> models.QuerySet:
        return self.annotate(
            expiration_time=ExpressionWrapper(
                F("created") + F("payment_window"), output_field=models.DateTimeField()
            )
        )

    def expired(self, at: Optional[datetime.datetime] = None) -> models.QuerySet:
        date_value = at or timezone.now()
        return self.with_expiration().filter(expiration_time__lt=date_value)

    def available(self, at: Optional[datetime.datetime] = None) -> models.QuerySet:
        date_value = at or timezone.now()
        return self.with_expiration().filter(
            created__lte=date_value, expiration_time__gte=date_value
        )


class Deposit(TimeStampedModel):
    STATUS = DEPOSIT_STATUS

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    session_key = models.SlugField(null=True)
    currency = models.ForeignKey(EthereumToken, on_delete=models.PROTECT)
    objects = DepositQuerySet.as_manager()

    @property
    def payments(self):
        return Payment.objects.filter(route__deposit=self).select_subclasses()

    @property
    def confirmed_payments(self):
        return self.payments.filter(confirmation__isnull=False)

    @property
    def total_transferred(self):
        return self.payments.aggregate(total=Sum("amount")).get("total") or 0

    @property
    def total_confirmed(self):
        return self.confirmed_payments.aggregate(total=Sum("amount")).get("total") or 0

    @property
    def is_expired(self):
        return all([route.is_expired for route in self.routes.select_subclasses()])

    @property
    def status(self):
        return self.STATUS.expired if self.is_expired else self.STATUS.open


class PaymentOrder(Deposit, EthereumTokenValueModel):
    objects = PaymentOrderQuerySet.as_manager()

    @property
    def due_amount(self):
        return max(0, self.amount - self.total_transferred)

    @property
    def is_paid(self):
        return self.due_amount <= 0

    @property
    def is_confirmed(self):
        return self.is_paid and self.total_confirmed >= self.amount

    @property
    def is_expired(self):
        return all([route.is_expired for route in self.routes.select_subclasses()])

    @property
    def status(self):
        if self.is_confirmed:
            return self.STATUS.confirmed
        elif self.is_paid:
            return self.STATUS.paid
        elif self.is_expired:
            return self.STATUS.expired
        else:
            return self.STATUS.open


class PaymentRoute(TimeStampedModel):
    NAME: Optional[str] = None

    deposit = models.ForeignKey(Deposit, on_delete=models.CASCADE, related_name="routes")
    objects = InheritanceManager()

    @property
    def name(self):
        return self.get_route_name()

    def get_route_name(self):
        if not self.NAME:
            route = PaymentRoute.objects.get_subclass(id=self.id)
            return route.NAME
        return self.NAME


class InternalPaymentRoute(PaymentRoute):
    NAME = "internal"


class BlockchainPaymentRoute(PaymentRoute):
    NAME = "blockchain"

    account = models.ForeignKey(
        BaseEthereumAccount, on_delete=models.CASCADE, related_name="blockchain_routes"
    )
    payment_window = IntegerRangeField()
    chain = models.ForeignKey(Chain, on_delete=models.CASCADE)
    objects = BlockchainRouteQuerySet.as_manager()

    @property
    def start_block_number(self):
        return self.payment_window.lower

    @property
    def expiration_block_number(self):
        return self.payment_window.upper

    @property
    def is_expired(self):
        return self.chain.highest_block > self.expiration_block_number

    @staticmethod
    def calculate_payment_window(chain):
        if not chain.synced:
            raise ValueError("Chain is not synced")

        current = chain.highest_block
        return (current, current + app_settings.Payment.blockchain_route_lifetime)


class RaidenPaymentRoute(PaymentRoute):
    NAME = "raiden"

    payment_window = models.DurationField(default=calculate_raiden_payment_window)
    raiden = models.ForeignKey(Raiden, on_delete=models.CASCADE, related_name="payment_routes")
    identifier = models.BigIntegerField(default=generate_payment_order_id, unique=True)

    objects = RaidenRouteQuerySet.as_manager()

    @property
    def is_expired(self):
        return self.expiration_time < timezone.now()

    @property
    def expiration_time(self):
        return self.created + self.payment_window

    @staticmethod
    def calculate_payment_window():
        return calculate_raiden_payment_window()

    class Meta:
        unique_together = ("raiden", "identifier")


class Payment(TimeStampedModel, EthereumTokenValueModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    route = models.ForeignKey(PaymentRoute, on_delete=models.PROTECT)
    objects = InheritanceManager()

    @property
    def is_confirmed(self):
        return hasattr(self, "confirmation")


class InternalPayment(Payment):
    memo = models.TextField(null=True, blank=True)

    @property
    def identifier(self):
        return str(self.id)


class BlockchainPayment(Payment):
    transaction = models.OneToOneField(Transaction, unique=True, on_delete=models.CASCADE)

    @property
    def identifier(self):
        return str(self.transaction.hash)


class RaidenPayment(Payment):
    payment = models.OneToOneField(RaidenPaymentEvent, unique=True, on_delete=models.CASCADE)

    @property
    def identifier(self):
        return f"{self.payment.identifier}-{self.id}"


class PaymentConfirmation(TimeStampedModel):
    payment = models.OneToOneField(Payment, on_delete=models.CASCADE, related_name="confirmation")


__all__ = [
    "Deposit",
    "PaymentOrder",
    "PaymentRoute",
    "InternalPaymentRoute",
    "BlockchainPaymentRoute",
    "RaidenPaymentRoute",
    "Payment",
    "InternalPayment",
    "BlockchainPayment",
    "RaidenPayment",
    "PaymentConfirmation",
]



from decimal import Decimal
from unittest.mock import MagicMock

from hub20.apps.blockchain.models import Transaction
from hub20.apps.core.choices import PAYMENT_NETWORKS
from hub20.apps.core.models import (
    BlockchainTransferExecutor,
    RaidenTransferExecutor,
    Transfer,
    TransferReceipt,
)
from hub20.apps.ethereum_money.factories import Erc20TransferFactory, EtherAmountFactory
from hub20.apps.ethereum_money.models import EthereumTokenAmount
from hub20.apps.ethereum_money.signals import outgoing_transfer_mined
from hub20.apps.raiden.factories import ChannelFactory, PaymentEventFactory


class MockBlockchainTransferExecutor(BlockchainTransferExecutor):
    def __init__(self, account, w3=None):
        self.account = account
        self.w3 = MagicMock()

    def execute(self, transfer: Transfer):
        tx = Erc20TransferFactory(
            recipient=transfer.address,
            from_address=self.account.address,
            amount=transfer.as_token_amount,
        )
        TransferReceipt.objects.create(
            transfer=transfer, network=PAYMENT_NETWORKS.blockchain, identifier=tx.hash
        )

        outgoing_transfer_mined.send(
            sender=Transaction,
            account=self.account,
            transaction=tx,
            amount=transfer.as_token_amount,
            address=transfer.address,
        )

    @classmethod
    def estimate_transfer_fees(cls, *args, **kw):
        return mock_fee_estimation()


class MockRaidenTransferExecutor(RaidenTransferExecutor):
    def execute(self, transfer: Transfer):
        identifier = self._ensure_valid_identifier(transfer.identifier)
        TransferReceipt.objects.create(
            transfer=transfer, network=PAYMENT_NETWORKS.raiden, identifier=identifier
        )
        channel = ChannelFactory(
            raiden=self.raiden, token_network__token=transfer.currency, balance=transfer.amount
        )
        params = dict(
            amount=transfer.amount,
            channel=channel,
            sender_address=self.raiden.address,
            receiver_address=transfer.address,
        )

        if identifier:
            params.update(dict(identifier=identifier))

        payment = PaymentEventFactory(**params)
        return str(payment.identifier)

    def transfer(self, *args, **kw):
        pass

    def __init__(self, account):
        self.raiden = account


def mock_fee_estimation() -> EthereumTokenAmount:
    return EtherAmountFactory(amount=Decimal("0.001"))



from typing import List

import factory
from hexbytes import HexBytes

from hub20.apps.blockchain.constants import ERC20_TRANSFER_TOPIC
from hub20.apps.blockchain.factories.providers import EthereumProvider
from hub20.apps.blockchain.tests.mocks import (
    TransactionDataMock,
    TransactionReceiptDataMock,
    Web3DataMock,
    Web3Model,
)
from hub20.apps.ethereum_money.client import encode_transfer_data
from hub20.apps.ethereum_money.factories import Erc20TokenAmountFactory, EtherAmountFactory

factory.Faker.add_provider(EthereumProvider)


def pad_hex(value: int, number_bytes: int) -> HexBytes:
    hex_value = hex(value)
    return HexBytes("0x" + hex_value[2:].zfill(number_bytes))


def pad_address(address: str) -> HexBytes:
    return HexBytes(address.replace("0x", "0x000000000000000000000000"))


def make_transfer_logs(tx_receipt_mock) -> List[Web3Model]:
    return [
        Web3Model(
            address=tx_receipt_mock.to,
            topics=[
                HexBytes(ERC20_TRANSFER_TOPIC),
                pad_address(tx_receipt_mock.from_address),
                pad_address(tx_receipt_mock.recipient),
            ],
            logIndex=0,
            blockNumber=tx_receipt_mock.blockNumber,
            blockHash=tx_receipt_mock.blockHash,
            transactionHash=tx_receipt_mock.hash,
            transactionIndex=0,
            data=pad_hex(tx_receipt_mock.amount.as_wei, 64),
        )
    ]


class EtherTransferDataMock(TransactionDataMock):
    to = factory.LazyAttribute(lambda obj: obj.recipient)
    gas = 21000
    value = factory.LazyAttribute(lambda obj: obj.amount.as_wei)

    class Params:
        recipient = factory.Faker("ethereum_address")
        amount = factory.SubFactory(EtherAmountFactory)


class Erc20TransferDataMock(TransactionDataMock):
    from_address = factory.Faker("ethereum_address")
    to = factory.LazyAttribute(lambda obj: obj.amount.currency.address)
    input = factory.LazyAttribute(lambda obj: encode_transfer_data(obj.recipient, obj.amount))

    class Params:
        recipient = factory.Faker("ethereum_address")
        amount = factory.SubFactory(Erc20TokenAmountFactory)


class EtherTransferReceiptMock(TransactionReceiptDataMock):
    to = factory.LazyAttribute(lambda obj: obj.recipient)
    value = factory.LazyAttribute(lambda obj: obj.amount.as_wei)

    class Params:
        recipient = factory.Faker("ethereum_address")
        amount = factory.SubFactory(Erc20TokenAmountFactory)


class Erc20TransferReceiptMock(TransactionReceiptDataMock):
    to = factory.LazyAttribute(lambda obj: obj.amount.currency.address)
    logs = factory.LazyAttribute(make_transfer_logs)

    class Params:
        recipient = factory.Faker("ethereum_address")
        amount = factory.SubFactory(Erc20TokenAmountFactory)


class Erc20LogFilterMock(Web3DataMock):
    args = factory.LazyAttribute(
        lambda obj: Web3Model(_from=obj.sender, _to=obj.recipient, _value=obj.amount.as_wei)
    )
    event = "Transfer"
    logIndex = 0
    transactionIndex = factory.fuzzy.FuzzyInteger(0, 50)
    transactionHash = factory.Faker("hex64")
    address = factory.LazyAttribute(lambda obj: obj.amount.currency.address)
    blockHash = factory.Faker("hex64")
    blockNumber = factory.Sequence(lambda n: n)

    class Params:
        recipient = factory.Faker("ethereum_address")
        sender = factory.Faker("ethereum_address")
        amount = factory.SubFactory(Erc20TokenAmountFactory)



# Generated by Django 4.0 on 2022-01-29 00:31

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import hub20.apps.blockchain.fields
import hub20.apps.ethereum_money.fields
import model_utils.fields
import taggit.managers
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('taggit', '0004_alter_taggeditem_content_type_alter_taggeditem_tag'),
        ('auth', '0012_alter_user_first_name_max_length'),
        ('blockchain', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='EthereumToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('address', hub20.apps.blockchain.fields.EthereumAddressField(default='0x0000000000000000000000000000000000000000')),
                ('symbol', models.CharField(max_length=20)),
                ('name', models.CharField(max_length=500)),
                ('decimals', models.PositiveIntegerField(default=18)),
                ('logoURI', hub20.apps.ethereum_money.fields.TokenlistStandardURLField(blank=True, max_length=512, null=True)),
                ('is_listed', models.BooleanField(default=False)),
                ('chain', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tokens', to='blockchain.chain')),
            ],
            options={
                'unique_together': {('chain', 'address')},
            },
        ),
        migrations.CreateModel(
            name='UserTokenList',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=64)),
                ('description', models.TextField(null=True)),
                ('keywords', taggit.managers.TaggableManager(help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags')),
                ('tokens', models.ManyToManyField(related_name='%(app_label)s_%(class)s_tokenlists', to='ethereum_money.EthereumToken')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='token_lists', to='auth.user')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='StableTokenPair',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('algorithmic_peg', models.BooleanField(default=True)),
                ('currency', models.CharField(choices=[('AED', 'United Arab Emirates dirham'), ('AFN', 'Afghan afghani'), ('ALL', 'Albanian lek'), ('AMD', 'Armenian dram'), ('ANG', 'Netherlands Antillean guilder'), ('AOA', 'Angolan kwanza'), ('ARS', 'Argentine peso'), ('AUD', 'Australian dollar'), ('AWG', 'Aruban florin'), ('AZN', 'Azerbaijani manat'), ('BAM', 'Bosnia and Herzegovina convertible mark'), ('BBD', 'Barbados dollar'), ('BDT', 'Bangladeshi taka'), ('BGN', 'Bulgarian lev'), ('BHD', 'Bahraini dinar'), ('BIF', 'Burundian franc'), ('BMD', 'Bermudian dollar'), ('BND', 'Brunei dollar'), ('BOB', 'Boliviano'), ('BOV', 'Bolivian Mvdol (funds code)'), ('BRL', 'Brazilian real'), ('BSD', 'Bahamian dollar'), ('BTN', 'Bhutanese ngultrum'), ('BWP', 'Botswana pula'), ('BYN', 'Belarusian ruble'), ('BZD', 'Belize dollar'), ('CAD', 'Canadian dollar'), ('CDF', 'Congolese franc'), ('CHE', 'WIR euro (complementary currency)'), ('CHF', 'Swiss franc'), ('CHW', 'WIR franc (complementary currency)'), ('CLF', 'Unidad de Fomento (funds code)'), ('CLP', 'Chilean peso'), ('CNY', 'Chinese yuan[8]'), ('COP', 'Colombian peso'), ('COU', 'Unidad de Valor Real (UVR) (funds code)[9]'), ('CRC', 'Costa Rican colon'), ('CUC', 'Cuban convertible peso'), ('CUP', 'Cuban peso'), ('CVE', 'Cape Verdean escudo'), ('CZK', 'Czech koruna'), ('DJF', 'Djiboutian franc'), ('DKK', 'Danish krone'), ('DOP', 'Dominican peso'), ('DZD', 'Algerian dinar'), ('EGP', 'Egyptian pound'), ('ERN', 'Eritrean nakfa'), ('ETB', 'Ethiopian birr'), ('EUR', 'Euro'), ('FJD', 'Fiji dollar'), ('FKP', 'Falkland Islands pound'), ('GBP', 'Pound sterling'), ('GEL', 'Georgian lari'), ('GHS', 'Ghanaian cedi'), ('GIP', 'Gibraltar pound'), ('GMD', 'Gambian dalasi'), ('GNF', 'Guinean franc'), ('GTQ', 'Guatemalan quetzal'), ('GYD', 'Guyanese dollar'), ('HKD', 'Hong Kong dollar'), ('HNL', 'Honduran lempira'), ('HRK', 'Croatian kuna'), ('HTG', 'Haitian gourde'), ('HUF', 'Hungarian forint'), ('IDR', 'Indonesian rupiah'), ('ILS', 'Israeli new shekel'), ('INR', 'Indian rupee'), ('IQD', 'Iraqi dinar'), ('IRR', 'Iranian rial'), ('ISK', 'Icelandic króna (plural: krónur)'), ('JMD', 'Jamaican dollar'), ('JOD', 'Jordanian dinar'), ('JPY', 'Japanese yen'), ('KES', 'Kenyan shilling'), ('KGS', 'Kyrgyzstani som'), ('KHR', 'Cambodian riel'), ('KMF', 'Comoro franc'), ('KPW', 'North Korean won'), ('KRW', 'South Korean won'), ('KWD', 'Kuwaiti dinar'), ('KYD', 'Cayman Islands dollar'), ('KZT', 'Kazakhstani tenge'), ('LAK', 'Lao kip'), ('LBP', 'Lebanese pound'), ('LKR', 'Sri Lankan rupee'), ('LRD', 'Liberian dollar'), ('LSL', 'Lesotho loti'), ('LYD', 'Libyan dinar'), ('MAD', 'Moroccan dirham'), ('MDL', 'Moldovan leu'), ('MGA', 'Malagasy ariary'), ('MKD', 'Macedonian denar'), ('MMK', 'Myanmar kyat'), ('MNT', 'Mongolian tögrög'), ('MOP', 'Macanese pataca'), ('MRU', 'Mauritanian ouguiya'), ('MUR', 'Mauritian rupee'), ('MVR', 'Maldivian rufiyaa'), ('MWK', 'Malawian kwacha'), ('MXN', 'Mexican peso'), ('MXV', 'Mexican Unidad de Inversion (UDI) (funds code)'), ('MYR', 'Malaysian ringgit'), ('MZN', 'Mozambican metical'), ('NAD', 'Namibian dollar'), ('NGN', 'Nigerian naira'), ('NIO', 'Nicaraguan córdoba'), ('NOK', 'Norwegian krone'), ('NPR', 'Nepalese rupee'), ('NZD', 'New Zealand dollar'), ('OMR', 'Omani rial'), ('PAB', 'Panamanian balboa'), ('PEN', 'Peruvian sol'), ('PGK', 'Papua New Guinean kina'), ('PHP', 'Philippine peso[13]'), ('PKR', 'Pakistani rupee'), ('PLN', 'Polish złoty'), ('PYG', 'Paraguayan guaraní'), ('QAR', 'Qatari riyal'), ('RON', 'Romanian leu'), ('RSD', 'Serbian dinar'), ('RUB', 'Russian ruble'), ('RWF', 'Rwandan franc'), ('SAR', 'Saudi riyal'), ('SBD', 'Solomon Islands dollar'), ('SCR', 'Seychelles rupee'), ('SDG', 'Sudanese pound'), ('SEK', 'Swedish krona (plural: kronor)'), ('SGD', 'Singapore dollar'), ('SHP', 'Saint Helena pound'), ('SLL', 'Sierra Leonean leone'), ('SOS', 'Somali shilling'), ('SRD', 'Surinamese dollar'), ('SSP', 'South Sudanese pound'), ('STN', 'São Tomé and Príncipe dobra'), ('SVC', 'Salvadoran colón'), ('SYP', 'Syrian pound'), ('SZL', 'Swazi lilangeni'), ('THB', 'Thai baht'), ('TJS', 'Tajikistani somoni'), ('TMT', 'Turkmenistan manat'), ('TND', 'Tunisian dinar'), ('TOP', 'Tongan paʻanga'), ('TRY', 'Turkish lira'), ('TTD', 'Trinidad and Tobago dollar'), ('TWD', 'New Taiwan dollar'), ('TZS', 'Tanzanian shilling'), ('UAH', 'Ukrainian hryvnia'), ('UGX', 'Ugandan shilling'), ('USD', 'United States dollar'), ('USN', 'United States dollar (next day) (funds code)'), ('UYI', 'Uruguay Peso en Unidades Indexadas (URUIURUI) (funds code)'), ('UYU', 'Uruguayan peso'), ('UYW', 'Unidad previsional[15]'), ('UZS', 'Uzbekistan som'), ('VED', 'Venezuelan bolívar digital[16]'), ('VES', 'Venezuelan bolívar soberano[13]'), ('VND', 'Vietnamese đồng'), ('VUV', 'Vanuatu vatu'), ('WST', 'Samoan tala'), ('XAF', 'CFA franc BEAC'), ('XAG', 'Silver (one troy ounce)'), ('XAU', 'Gold (one troy ounce)'), ('XBA', 'European Composite Unit (EURCO) (bond market unit)'), ('XBB', 'European Monetary Unit (E.M.U.-6) (bond market unit)'), ('XBC', 'European Unit of Account 9 (E.U.A.-9) (bond market unit)'), ('XBD', 'European Unit of Account 17 (E.U.A.-17) (bond market unit)'), ('XCD', 'East Caribbean dollar'), ('XDR', 'Special drawing rights'), ('XOF', 'CFA franc BCEAO'), ('XPD', 'Palladium (one troy ounce)'), ('XPF', 'CFP franc (franc Pacifique)'), ('XPT', 'Platinum (one troy ounce)'), ('XSU', 'SUCRE'), ('XTS', 'Code reserved for testing'), ('XUA', 'ADB Unit of Account'), ('XXX', 'No currency'), ('YER', 'Yemeni rial'), ('ZAR', 'South African rand'), ('ZMW', 'Zambian kwacha'), ('ZWL', 'Zimbabwean dollar')], max_length=3)),
                ('token', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='stable_pair', to='ethereum_money.ethereumtoken')),
            ],
        ),
        migrations.CreateModel(
            name='WrappedToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('wrapped', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='wrapping_tokens', to='ethereum_money.ethereumtoken')),
                ('wrapper', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='ethereum_money.ethereumtoken')),
            ],
            options={
                'unique_together': {('wrapped', 'wrapper')},
            },
        ),
        migrations.CreateModel(
            name='TokenList',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=64)),
                ('description', models.TextField(null=True)),
                ('url', hub20.apps.ethereum_money.fields.TokenlistStandardURLField()),
                ('version', models.CharField(max_length=32)),
                ('keywords', taggit.managers.TaggableManager(help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags')),
                ('tokens', models.ManyToManyField(related_name='%(app_label)s_%(class)s_tokenlists', to='ethereum_money.EthereumToken')),
            ],
            options={
                'unique_together': {('url', 'version')},
            },
        ),
    ]



import asyncio
import logging

import celery_pubsub
from asgiref.sync import sync_to_async
from django.core.cache import cache

from hub20.apps.blockchain.client import BLOCK_CREATION_INTERVAL, make_web3
from hub20.apps.blockchain.models import Web3Provider
from hub20.apps.core.models import BlockchainPaymentRoute
from hub20.apps.ethereum_money.abi import EIP20_ABI
from hub20.apps.ethereum_money.models import EthereumToken

logger = logging.getLogger(__name__)


CACHE_KEY = "TRANSACTIONS_FOR_OPEN_ROUTES"


async def pending_token_transfers():
    logger.debug("Checking for pending token transfers")
    while True:
        open_routes = await sync_to_async(list)(
            BlockchainPaymentRoute.objects.open().select_related(
                "deposit",
                "deposit__currency",
                "deposit__currency__chain",
                "account",
            )
        )

        for route in open_routes:
            logger.info(f"Checking for pending token transfers for payment {route.deposit_id}")
            token: EthereumToken = route.deposit.currency

            # We are only concerned here about ERC20
            # tokens. ETH transfers are detected directly
            # by the blockchain listeners
            if not token.is_ERC20:
                continue

            provider = await sync_to_async(Web3Provider.active.filter(chain=token.chain).first)()

            if not provider:
                logger.warning(
                    f"Route {route} is open but not provider available to check for payments"
                )
                continue

            w3 = make_web3(provider=provider)
            contract = w3.eth.contract(abi=EIP20_ABI, address=token.address)

            try:
                for transfer_event in contract.events.Transfer().getLogs(
                    {"fromBlock": "pending", "argument_filters": {"_to": route.account.address}}
                ):
                    tx_hash = transfer_event.transactionHash.hex()

                    key = f"{CACHE_KEY}:{tx_hash}"

                    if await sync_to_async(cache.get)(key):
                        logger.debug(f"Event for tx {tx_hash} has already been published")
                        continue

                    tx_data = w3.eth.get_transaction(transfer_event.transactionHash)

                    await sync_to_async(celery_pubsub.publish)(
                        "blockchain.broadcast.event",
                        chain_id=w3.eth.chain_id,
                        transaction_data=tx_data,
                        event=transfer_event,
                    )
                    await sync_to_async(cache.set)(key, True, timeout=BLOCK_CREATION_INTERVAL * 2)
            except ValueError:
                logger.warning(f"Can not get transfer logs from {provider.hostname}")
        else:
            await asyncio.sleep(1)



import factory
from django.contrib.auth import get_user_model

User = get_user_model()


class UserFactory(factory.django.DjangoModelFactory):
    username = factory.Sequence(lambda n: f"test-user-{n:03}")
    is_active = True
    is_staff = False
    is_superuser = False

    class Meta:
        model = User


__all__ = ["UserFactory"]



import logging

from django.utils.translation import gettext_lazy as _
from ethereum.utils import checksum_encode, normalize_address
from hexbytes import HexBytes
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from .analytics import estimate_gas_price
from .constants import (
    SIGNATURE_R_MAX_VALUE,
    SIGNATURE_R_MIN_VALUE,
    SIGNATURE_S_MAX_VALUE,
    SIGNATURE_S_MIN_VALUE,
    SIGNATURE_V_MAX_VALUE,
    SIGNATURE_V_MIN_VALUE,
)
from .models import Chain

logger = logging.getLogger(__name__)


class ChainSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="blockchain:chain-detail")
    token = serializers.CharField(source="native_token.symbol", read_only=True)
    explorers = serializers.SerializerMethodField()
    status = serializers.HyperlinkedIdentityField(view_name="blockchain:chain-status")

    def get_explorers(self, obj):
        return obj.explorers.values_list("url", flat=True)

    class Meta:
        model = Chain
        fields = read_only_fields = (
            "url",
            "id",
            "name",
            "token",
            "explorers",
            "status",
        )


class ChainStatusSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="blockchain:chain-status")
    chain = serializers.HyperlinkedIdentityField(view_name="blockchain:chain-detail")
    height = serializers.IntegerField(source="highest_block", read_only=True)
    online = serializers.BooleanField(source="provider.connected")
    synced = serializers.BooleanField(source="provider.synced")
    gas_price_estimate = serializers.SerializerMethodField()

    def get_gas_price_estimate(self, obj):
        return estimate_gas_price(obj.id)

    class Meta:
        model = Chain
        fields = read_only_fields = (
            "url",
            "chain",
            "height",
            "online",
            "synced",
            "gas_price_estimate",
        )


# ================================================ #
#                Custom Fields
# ================================================ #
class EthereumAddressField(serializers.Field):
    """
    Ethereum address checksumed
    https://github.com/ethereum/EIPs/blob/master/EIPS/eip-55.md
    """

    def __init__(self, allow_zero_address=False, allow_sentinel_address=False, **kwargs):
        self.allow_zero_address = allow_zero_address
        self.allow_sentinel_address = allow_sentinel_address
        super().__init__(**kwargs)

    def to_representation(self, obj):
        return obj

    def to_internal_value(self, data):
        try:
            if int(data, 16) == 0 and not self.allow_zero_address:
                raise ValidationError("0x0 address is not allowed")
            elif int(data, 16) == 1 and not self.allow_sentinel_address:
                raise ValidationError("0x1 address is not allowed")

            return checksum_encode(normalize_address(int(data, 16)))
        except Exception:
            raise ValidationError("Address %s is not valid" % data)


class HexadecimalField(serializers.Field):
    """
    Serializes hexadecimal values starting by `0x`. Empty values should be None or just `0x`.
    """

    default_error_messages = {
        "invalid": _("{value} is not an hexadecimal value."),
        "blank": _("This field may not be blank."),
        "max_length": _(
            "Ensure this field has no more than {max_length} hexadecimal chars (not counting 0x)."
        ),
        "min_length": _(
            "Ensure this field has at least {min_length} hexadecimal chars (not counting 0x)."
        ),
    }

    def __init__(self, **kwargs):
        self.allow_blank = kwargs.pop("allow_blank", False)
        self.max_length = kwargs.pop("max_length", None)
        self.min_length = kwargs.pop("min_length", None)
        super().__init__(**kwargs)

    def to_representation(self, obj):
        if not obj:
            return "0x"

        # We can get another types like `memoryview` from django models.
        # `to_internal_value` is not used when you provide an object
        # instead of a json using `data`. Make sure everything is HexBytes.
        if hasattr(obj, "hex"):
            obj = HexBytes(obj.hex())
        elif not isinstance(obj, HexBytes):
            obj = HexBytes(obj)
        return obj.hex()

    def to_internal_value(self, data):
        if isinstance(data, (bytes, memoryview)):
            data = data.hex()

        data = data.strip()  # Trim spaces
        if data.startswith("0x"):  # Remove 0x prefix
            data = data[2:]

        if not data:
            if self.allow_blank:
                return None
            else:
                self.fail("blank")

        data_len = len(data)
        if self.min_length and data_len < self.min_length:
            self.fail("min_length", min_length=data_len)
        elif self.max_length and data_len > self.max_length:
            self.fail("max_length", max_length=data_len)

        try:
            return HexBytes(data)
        except ValueError:
            self.fail("invalid", value=data)


class Sha3HashField(HexadecimalField):
    def __init__(self, **kwargs):
        kwargs["max_length"] = 64
        kwargs["min_length"] = 64
        super().__init__(**kwargs)


# ================================================ #
#                Base Serializers
# ================================================ #
class SignatureSerializer(serializers.Serializer):
    v = serializers.IntegerField(min_value=SIGNATURE_V_MIN_VALUE, max_value=SIGNATURE_V_MAX_VALUE)
    r = serializers.IntegerField(min_value=SIGNATURE_R_MIN_VALUE, max_value=SIGNATURE_R_MAX_VALUE)
    s = serializers.IntegerField(min_value=SIGNATURE_S_MIN_VALUE, max_value=SIGNATURE_S_MAX_VALUE)


class TransactionSerializer(serializers.Serializer):
    from_ = EthereumAddressField()
    value = serializers.IntegerField(min_value=0)
    data = HexadecimalField()
    gas = serializers.IntegerField(min_value=0)
    gas_price = serializers.IntegerField(min_value=0)
    nonce = serializers.IntegerField(min_value=0)

    def get_fields(self):
        result = super().get_fields()
        # Rename `from_` to `from`
        from_ = result.pop("from_")
        result["from"] = from_
        return result


class TransactionResponseSerializer(serializers.Serializer):
    """
    Use chars to avoid problems with big ints (i.e. JavaScript)
    """

    from_ = EthereumAddressField()
    value = serializers.IntegerField(min_value=0)
    data = serializers.CharField()
    gas = serializers.CharField()
    gas_price = serializers.CharField()
    nonce = serializers.IntegerField(min_value=0)

    def get_fields(self):
        result = super().get_fields()
        # Rename `from_` to `from`
        from_ = result.pop("from_")
        result["from"] = from_
        return result



from django.db.models.query import QuerySet
from rest_framework import generics, mixins, status, viewsets
from rest_framework.decorators import action
from rest_framework.permissions import IsAdminUser
from rest_framework.response import Response

from . import models, serializers


class BaseRaidenViewMixin:
    permission_classes = (IsAdminUser,)


class RaidenViewSet(
    BaseRaidenViewMixin, mixins.ListModelMixin, mixins.RetrieveModelMixin, viewsets.GenericViewSet
):
    serializer_class = serializers.RaidenSerializer
    queryset = models.Raiden.objects.all()

    @action(detail=True, methods=["get"])
    def status(self, request, pk=None):
        raiden = self.get_object()
        serializer = serializers.RaidenStatusSerializer(raiden, context={"request": request})
        return Response(serializer.data)


class ChannelViewMixin(BaseRaidenViewMixin):
    serializer_class = serializers.ChannelSerializer


class ChannelViewSet(
    ChannelViewMixin, mixins.ListModelMixin, mixins.RetrieveModelMixin, viewsets.GenericViewSet
):
    def get_queryset(self, *args, **kw):
        return models.Channel.objects.filter(raiden_id=self.kwargs["raiden_pk"])

    def get_object(self):
        return self.get_queryset().filter(pk=self.kwargs["pk"]).first()

    @action(
        detail=True,
        methods=["POST"],
        serializer_class=serializers.ChannelDepositSerializer,
    )
    def deposit(self, request, *args, **kw):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        else:
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    @action(
        detail=True,
        methods=["POST"],
        serializer_class=serializers.ChannelWithdrawSerializer,
    )
    def withdraw(self, request, *args, **kw):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        else:
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class ServiceDepositMixin(BaseRaidenViewMixin):
    serializer_class = serializers.ServiceDepositSerializer
    queryset = models.UserDepositContractOrder.objects.all()


class ServiceDepositListView(ServiceDepositMixin, generics.ListCreateAPIView):
    pass


class ServiceDepositDetailView(ServiceDepositMixin, generics.RetrieveAPIView):
    pass


class TokenNetworkViewMixin:
    permission_classes = (IsAdminUser,)
    serializer_class = serializers.TokenNetworkSerializer
    lookup_field = "address"
    lookup_url_kwarg = "address"
    queryset: QuerySet = models.TokenNetwork.objects.all()


class TokenNetworkViewSet(
    TokenNetworkViewMixin,
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    mixins.DestroyModelMixin,
    viewsets.GenericViewSet,
):
    def destroy(self, request, *args, **kw):
        raiden = models.Raiden.objects.first()

        if raiden:
            models.LeaveTokenNetworkOrder.objects.create(
                raiden=raiden, user=request.user, token_network=self.get_object()
            )
        return Response(status=status.HTTP_204_NO_CONTENT)

    @action(
        detail=True, methods=["post"], serializer_class=serializers.JoinTokenNetworkOrderSerializer
    )
    def join(self, request, address=None):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        else:
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)



import logging

from django.db.models.signals import post_save
from django.dispatch import receiver

from hub20.apps.blockchain.models import Transaction
from hub20.apps.core.models import BlockchainTransferExecution, RaidenTransferExecution, Transfer
from hub20.apps.ethereum_money.signals import outgoing_transfer_mined
from hub20.apps.raiden.models import Payment

logger = logging.getLogger(__name__)


@receiver(outgoing_transfer_mined, sender=Transaction)
def on_blockchain_transfer_mined_record_execution(sender, **kw):
    amount = kw["amount"]
    transaction = kw["transaction"]
    address = kw["address"]

    transfer = Transfer.processed.filter(
        amount=amount.amount,
        currency=amount.currency,
        address=address,
        receipt__identifier=transaction.hash,
    ).first()

    if transfer:
        BlockchainTransferExecution.objects.create(transfer=transfer, transaction=transaction)


@receiver(post_save, sender=Payment)
def on_raiden_payment_sent_record_execution(sender, **kw):
    if kw["created"]:
        payment = kw["instance"]
        transfer = Transfer.processed.filter(
            amount=payment.amount,
            currency=payment.token,
            address=payment.receiver_address,
            receipt__identifier=payment.identifier,
        ).first()

        if transfer:
            RaidenTransferExecution.objects.create(transfer=transfer, payment=payment)


__all__ = [
    "on_blockchain_transfer_mined_record_execution",
    "on_raiden_payment_sent_record_execution",
]



from django.test import TestCase
from django.urls import reverse
from rest_framework.test import APIClient

from hub20.apps.raiden import factories


class BaseRaidenAdminViewTestCase(TestCase):
    def setUp(self):
        self.admin = factories.AdminUserFactory()
        self.client = APIClient()
        self.client.force_authenticate(user=self.admin)
        self.raiden = factories.RaidenFactory()
        self.access_url = reverse("raiden-detail", kwargs={"pk": self.raiden.pk})


class RaidenNodeViewTestCase(BaseRaidenAdminViewTestCase):
    def test_anonymous_user_can_not_access_endpoints(self):
        client = APIClient()
        response = client.get(self.access_url)
        self.assertEqual(response.status_code, 401)

    def test_regular_user_can_not_access_endpoints(self):
        user = factories.UserFactory()
        client = APIClient()
        client.force_authenticate(user=user)
        response = client.get(self.access_url)
        self.assertEqual(response.status_code, 403)

    def test_admin_can_see_raiden_endpoint(self):
        response = self.client.get(self.access_url)
        self.assertEqual(response.status_code, 200)


class ChannelViewTestCase(BaseRaidenAdminViewTestCase):
    def setUp(self):
        self.channel = factories.ChannelFactory()

    def test_can_get_deposit_url(self):
        kwargs = {"raiden_pk": self.channel.raiden.id, "pk": self.channel.id}
        url = reverse("raiden-channels-deposit", kwargs=kwargs)
        self.assertTrue(url.endswith(f"channels/{self.channel.id}/deposit"))

    def test_can_get_withdrawal_url(self):
        kwargs = {"raiden_pk": self.channel.raiden.id, "pk": self.channel.id}
        url = reverse("raiden-channels-withdraw", kwargs=kwargs)
        self.assertTrue(url.endswith(f"channels/{self.channel.id}/withdraw"))


__all__ = ["RaidenNodeViewTestCase", "ChannelViewTestCase"]



from decimal import Decimal

import pytest
from asgiref.sync import sync_to_async
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.testing import WebsocketCommunicator
from django.core.asgi import get_asgi_application
from eth_utils import is_0x_prefixed
from web3 import Web3

from hub20.apps.blockchain.factories import BlockFactory
from hub20.apps.blockchain.models import BaseEthereumAccount, Chain
from hub20.apps.blockchain.tests.mocks import BlockMock
from hub20.apps.core import handlers
from hub20.apps.core.api import consumer_patterns
from hub20.apps.core.consumers import Events
from hub20.apps.core.factories import (
    CheckoutFactory,
    Erc20TokenBlockchainPaymentFactory,
    Erc20TokenPaymentOrderFactory,
)
from hub20.apps.core.middleware import TokenAuthMiddlewareStack
from hub20.apps.core.settings import app_settings
from hub20.apps.ethereum_money.abi import EIP20_ABI
from hub20.apps.ethereum_money.tests.mocks import Erc20TransferDataMock, Erc20TransferReceiptMock

application = ProtocolTypeRouter(
    {
        "http": get_asgi_application(),
        "websocket": TokenAuthMiddlewareStack(URLRouter(consumer_patterns)),
    }
)


def deposit_account(payment_request):
    return BaseEthereumAccount.objects.filter(blockchain_routes__deposit=payment_request).first()


def deposit_transaction_params(payment_request):
    route = payment_request.routes.select_subclasses().first()
    return dict(
        blockNumber=payment_request.currency.chain.highest_block,
        recipient=route.account.address,
        amount=payment_request.as_token_amount,
    )


def deposit_tx_data(deposit_tx_params):
    return Erc20TransferDataMock(**deposit_tx_params)


def deposit_tx_receipt(tx_data, tx_params):
    return Erc20TransferReceiptMock(
        hash=tx_data.hash,
        blockHash=tx_data.blockHash,
        from_address=tx_data["from"],
        transactionIndex=tx_data.transactionIndex,
        **tx_params,
    )


def deposit_block_data(tx_data):
    return BlockMock(
        hash=tx_data.blockHash,
        number=tx_data.blockNumber,
        transactions=[tx_data.hash],
    )


def erc20_deposit_transfer_events(tx_data, tx_params):
    w3 = Web3()
    tx_receipt = deposit_tx_receipt(tx_data, tx_params)
    contract = w3.eth.contract(address=tx_receipt.to, abi=EIP20_ABI)
    return contract.events.Transfer().processReceipt(tx_receipt)


@pytest.fixture
def session_events_communicator(client):
    communicator = WebsocketCommunicator(application, "events")
    communicator.scope["session"] = client.session

    return communicator


@pytest.fixture
def erc20_payment_request(client):
    return Erc20TokenPaymentOrderFactory(session_key=client.session.session_key)


@pytest.fixture
def erc20_blockchain_payment(erc20_payment_request):
    return Erc20TokenBlockchainPaymentFactory(route__deposit=erc20_payment_request)


@pytest.fixture
def checkout():
    checkout = CheckoutFactory()
    checkout.store.accepted_token_list.tokens.add(checkout.currency)
    return checkout


@pytest.fixture
def erc20_blockchain_checkout_payment(checkout):
    return Erc20TokenBlockchainPaymentFactory(
        currency=checkout.currency, amount=checkout.amount, route__deposit=checkout
    )


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_session_receives_token_deposit_received(
    session_events_communicator,
    erc20_blockchain_payment,
):

    ok, protocol_or_error = await session_events_communicator.connect()
    assert ok, "Failed to connect"

    await sync_to_async(handlers.on_blockchain_payment_received_send_notification)(
        sender=erc20_blockchain_payment.__class__, payment=erc20_blockchain_payment
    )

    messages = []
    while not await session_events_communicator.receive_nothing(timeout=0.25):
        messages.append(await session_events_communicator.receive_json_from())

    await session_events_communicator.disconnect()

    assert len(messages) != 0, "we should have received something here"
    payment_received_event = Events.BLOCKCHAIN_DEPOSIT_RECEIVED.value

    payment_received_messages = [msg for msg in messages if msg["event"] == payment_received_event]
    assert len(payment_received_messages) == 1, "we should have a payment received message"

    payment_message = payment_received_messages[0]
    assert "data" in payment_message

    payment_data = payment_message["data"]
    assert is_0x_prefixed(payment_data["transaction"])


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_checkout_receives_transaction_mined_notification(
    checkout, erc20_blockchain_checkout_payment
):
    communicator = WebsocketCommunicator(application, f"checkout/{checkout.id}")

    ok, protocol_or_error = await communicator.connect()
    assert ok, "Failed to connect"

    payment = erc20_blockchain_checkout_payment

    await sync_to_async(handlers.on_blockchain_payment_received_send_notification)(
        sender=payment.__class__, payment=payment
    )

    messages = []
    while not await communicator.receive_nothing(timeout=0.25):
        messages.append(await communicator.receive_json_from())

    await communicator.disconnect()

    assert len(messages) != 0, "we should have received something here"
    payment_mined_event = Events.BLOCKCHAIN_DEPOSIT_RECEIVED.value

    payment_messages = [msg for msg in messages if msg["event"] == payment_mined_event]
    assert len(payment_messages) == 1, "we should have received one payment received message"

    payment_data = payment_messages[0]
    assert is_0x_prefixed(payment_data["transaction"])
    assert is_0x_prefixed(payment_data["token"])
    assert Decimal(payment_data["amount"]) == checkout.amount, "payment amount does not match"


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_checkout_receives_transaction_broadcast_notification(
    checkout, erc20_blockchain_checkout_payment
):
    communicator = WebsocketCommunicator(application, f"checkout/{checkout.id}")

    ok, protocol_or_error = await communicator.connect()
    assert ok, "Failed to connect"

    account = await sync_to_async(deposit_account)(checkout)
    tx_params = await sync_to_async(deposit_transaction_params)(checkout)
    tx_data = deposit_tx_data(tx_params)

    await sync_to_async(
        handlers.on_incoming_transfer_broadcast_send_notification_to_open_checkouts
    )(
        sender=tx_data.__class__,
        account=account,
        amount=checkout.as_token_amount,
        transaction_data=tx_data,
    )

    messages = []
    while not await communicator.receive_nothing(timeout=0.25):
        messages.append(await communicator.receive_json_from())

    await communicator.disconnect()

    assert len(messages) != 0, "we should have received something here"
    payment_sent_event = Events.BLOCKCHAIN_DEPOSIT_BROADCAST.value

    payment_messages = [msg for msg in messages if msg["event"] == payment_sent_event]
    assert len(payment_messages) == 1, "we should have received one payment sent message"

    payment_data = payment_messages[0]
    assert is_0x_prefixed(payment_data["transaction"])
    assert is_0x_prefixed(payment_data["token"])
    assert Decimal(payment_data["amount"]) == checkout.amount, "payment amount does not match"


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_checkout_receives_confirmation_notification(
    checkout, erc20_blockchain_checkout_payment
):
    communicator = WebsocketCommunicator(application, f"checkout/{checkout.id}")

    ok, protocol_or_error = await communicator.connect()
    assert ok, "Failed to connect"

    tx_block_number = erc20_blockchain_checkout_payment.transaction.block.number
    confirmation_block_number = tx_block_number + app_settings.Payment.minimum_confirmations

    block = await sync_to_async(BlockFactory)(number=confirmation_block_number)
    block.chain.highest_block = confirmation_block_number

    await sync_to_async(handlers.on_chain_updated_check_payment_confirmations)(
        sender=Chain, instance=block.chain
    )

    messages = []
    while not await communicator.receive_nothing(timeout=0.25):
        messages.append(await communicator.receive_json_from())

    await communicator.disconnect()

    assert len(messages) != 0, "we should have received something here"
    payment_confirmed_event = Events.BLOCKCHAIN_DEPOSIT_CONFIRMED.value

    payment_messages = [msg for msg in messages if msg["event"] == payment_confirmed_event]
    assert len(payment_messages) == 1, "we should have received one payment confirmed message"


__all__ = [
    "test_session_receives_token_deposit_received",
    "test_checkout_receives_transaction_mined_notification",
    "test_checkout_receives_transaction_broadcast_notification",
    "test_checkout_receives_confirmation_notification",
]



from model_utils import Choices

DEPOSIT_STATUS = Choices("open", "paid", "confirmed", "expired")
TRANSFER_STATUS = Choices("scheduled", "processed", "failed", "canceled", "executed")
PAYMENT_NETWORKS = Choices("blockchain", "raiden", "internal")
